#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Prioritaetswarteschlange mittels eines binaeren Heaps unter Nutzung von Arrays
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Datentypen nutzen!

	Foliensatz: 04-Prioritaetswarteschlangen
	Ab Folie 3
"""

include("./../Utils/key-mapping.jl")

# Datentypen
struct BinaererHeapPQ
	internal_repr :: Array{Any}
	key :: Function 
	set_key :: Function
	mapping :: KeyMapping # Array[key(element)] = index in internal_repr
end

# Funktionen extern
function build(elements, key::Function, set_key::Function)
	elements = convert(Vector{Any}, elements)

	mapping = KeyMapping()
	for (i, e) in enumerate(elements)
		insert(key(e), i, mapping)
	end
	internal_set_key = function (element, key_value) 
		# Aktualisiert den Key sowohl im Mapping als auch
		# ueber set_key() im Element selbst
		update_key(key(element), key_value, mapping)
		set_key(element, key_value)
	end

	pq = BinaererHeapPQ(
		elements,
		key, internal_set_key,
		mapping
	)

	H = pq.internal_repr
	n = length(H)
	for i = parent(n):-1:1
		sift_down(i, pq)
	end
	return pq
end

function insert(e, pq::BinaererHeapPQ)
	H = pq.internal_repr
	n = length(H) + 1

	insert(pq.key(e), n, pq.mapping)

	insert!(H, n, e)
	sift_up(n, pq)

	nothing
end 

function min_element(pq::BinaererHeapPQ)
	return pq.internal_repr[1]
end 

function delete_min(pq::BinaererHeapPQ)
	H = pq.internal_repr
	n = length(H)
	e = H[1]

	H[1] = H[n]	
	update(pq.key(H[1]), 1, pq.mapping)

	deleteat!(H, n)
	delete(pq.key(e), pq.mapping)

	sift_down(1, pq)

	return e
end

function empty_pq(pq::BinaererHeapPQ)
	return length(pq.internal_repr) == 0
end

function delete(e::Any, pq::BinaererHeapPQ)
	H = pq.internal_repr
	n = length(H)

	i = lookup(pq.key(e), pq.mapping)
	if !isnothing(i)
		H[i] = H[n]	
		update(pq.key(H[i]), i, pq.mapping)

		deleteat!(H, n)
		delete(pq.key(e), pq.mapping)

		sift_down(i, pq)
	end
	
	nothing
end

function decrease_key(e::Any, pq::BinaererHeapPQ, delta::Int)
	H = pq.internal_repr

	i = lookup(pq.key(e), pq.mapping)
	if !isnothing(i)
		pq.set_key(e, pq.key(e) - delta)
		sift_up(i, pq)
	end 

	nothing
end

function merge(pq::BinaererHeapPQ, pq_::BinaererHeapPQ)
	H = pq.internal_repr
	n = length(H)
	merge(pq.mapping, pq_.mapping, (i) -> n+i)

	append!(H, pq_.internal_repr)
	n = length(H)

	for i = parent(n):-1:1
		sift_down(i, pq)
	end

	nothing
end

# Funktionen intern 
function sift_up(i::Int, pq::BinaererHeapPQ)
	H = pq.internal_repr

	while i > 1 && pq.key(H[parent(i)] ) > pq.key(H[i])
		temp = H[i]
		H[i] = H[parent(i)]
		H[parent(i)] = temp

		update(pq.key(H[i]), i, pq.mapping)
		update(pq.key(H[parent(i)]), parent(i), pq.mapping)

		i = parent(i)
	end
end

function sift_down(i::Int, pq::BinaererHeapPQ)
	H = pq.internal_repr

	while !is_leaf(i, H)
		if !exists(right_child(i), H) 
			m = left_child(i)
		else 
			if pq.key(H[left_child(i)]) < pq.key(H[right_child(i)])
				m = left_child(i)
			else 
				m = right_child(i)
			end 
		end
		if pq.key(H[i]) <= pq.key(H[m])
			return # Invariante erfuellt
		else 
			temp = H[i]
			H[i] = H[m]
			H[m] = temp

			update(pq.key(H[i]), i, pq.mapping)
			update(pq.key(H[m]), m, pq.mapping)
		end
		i = m
	end
end

function parent(i)
    return i ÷ 2
end

function is_leaf(i, H)
    return 2 * i > length(H)
end

function exists(i, H)
    return i <= length(H)
end

function left_child(i)
    return 2 * i
end

function right_child(i)
    return 2 * i + 1
end

# Ausgabe als Baum mit durch Julia
function Base.show(io::IO, pq::BinaererHeapPQ)
	H = pq.internal_repr
	n = length(H)
	pad_size = 12

	println("BinaererHeapPQ mit ", n, " Elementen:")
	height = ceil(log2(n)) 
	width = Int(2^(height-1) * pad_size)
	for i = 1:height+1
		s = Int(2^(i-1))
		e = Int(2^i-1)
		pad = width ÷ (e - s + 1)
		for j = s:minimum([e, n])
			print(io, lpad(rpad(H[j], pad ÷ 2), pad))
		end
		println(io)
	end
	#println(io, "key(Element) => index: ", pq.mapping)
end

# Beispielaufrufe
# 	Beispieldatentyp
mutable struct Process 
	processID :: Int 
	priority :: Int
end
# 	Ausgabe 
function Base.show(io::IO, e::Process)
	print(io, "$(e.processID) (key=$(e.priority))")
end
# 	Erstellung von "Prozessen" nur durch Angaben der Prioritaet(en)
function Process(p::Int)
	return Process(p, p)
end
function Processes(ps...)
	es = Array{Process}(undef, length(ps))
	for (i,p) in enumerate(ps)
		es[i] = Process(p)
	end
	return es
end
# 	Funktion zum Holen und Setzen der Prioritaet
function get_process_key(e) 
	return e.priority
end
function set_process_key(e, priority) 
	e.priority = priority
end

# Aufrufe
my_es = Processes(9, 11, 3, 1, 7, 5)
my_pq = build(my_es, get_process_key, set_process_key)
println(my_pq)
#	weitere PQ erstellen, fuer merge
my_other_es = Processes(10, 2, 4, 8, 6, 12)
my_other_pq = build(my_other_es, get_process_key, set_process_key)
println(my_other_pq)
#	Merge der beiden PQs
merge(my_pq, my_other_pq)
println(my_pq)
# 	Minima loeschen
println(min_element(my_pq), ' ', delete_min(my_pq))
println(empty_pq(my_pq))
println(delete_min(my_pq), ' ', delete_min(my_pq))
println(my_pq)
#	Andere loeschen
println("Loesche: ", my_other_es[1])
delete(my_other_es[1], my_pq) 
println(my_pq)
println("Loesche: ", my_other_es[4])
delete(my_other_es[4], my_pq)
println(my_pq)
# 	Wieder einfuegen
my_e = Process(10)
println("Fuege ein: ", my_e)
insert(my_e, my_pq) 
println(my_pq)
#	Key verringern
println("Verringere Key von ", my_other_es[6], " um 9")
decrease_key(my_other_es[6], my_pq, 9)
println(my_pq)