#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Prioritaetswarteschlange mittels Binomial Heaps
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Datentypen nutzen!

	Foliensatz: 04-Prioritaetswarteschlangen
	Ab Folie 29
"""

# Linked List analog zu Foliensatz 03, Folie 54 laden
include("./../Utils/linked-list.jl")
include("./../Utils/key-mapping.jl")

# Datentypen
mutable struct BinomialHeapNode 
	value :: Any
	parent :: Union{BinomialHeapNode, Nothing}
	children :: Array{BinomialHeapNode}
end

mutable struct BinomialHeapPQ
	min :: Union{BinomialHeapNode, Nothing}
	roots :: LinkedList
	key :: Function 
	set_key :: Function
	mapping :: KeyMapping # Array[key(elment)] = PQ node
end

# Funktione zum Zugriff auf Attribute
function rank(node::BinomialHeapNode)
	return length(node.children)
end

function parent(node::BinomialHeapNode)
	return node.parent
end

function children(node::BinomialHeapNode)
	return node.children
end

# Funktionen extern
function build(elements, key::Function, set_key::Function)
	mapping = KeyMapping()
	internal_set_key = function (element, key_value) 
		# Aktualisiert den Key sowohl im Mapping als auch
		# ueber set_key() im Element selbst
		update_key(key(element), key_value, mapping)
		set_key(element, key_value)
	end

	pq = BinomialHeapPQ(
		nothing, make_list(),
		key, internal_set_key,
		mapping
	)
	for e in elements
		insert(e, pq)
	end
	return pq
end

function insert(e::Any, pq::BinomialHeapPQ)
	node = BinomialHeapNode(e, nothing, [])
	insert(pq.key(e), node, pq.mapping)
	insert(node, pq)
end
function insert(node::BinomialHeapNode, pq::BinomialHeapPQ)
	new_pq = BinomialHeapPQ(node, cons(node, make_list()), pq.key, pq.set_key, pq.mapping)
	merge(pq, new_pq)

	nothing
end

function min_element(pq::BinomialHeapPQ)
	return min_node(pq).value
end
function min_node(pq::BinomialHeapPQ)
	return pq.min
end

function delete_min(pq::BinomialHeapPQ)
	node = min_node(pq)
	delete(pq.key(node.value), pq.mapping)

	delete(node, pq.roots)
	for c in children(node)
		c.parent = nothing
		insert(c, pq.roots)
	end
	merge(pq, BinomialHeapPQ(nothing, make_list(), pq.key, pq.set_key, pq.mapping))

	pq.min = nothing 
	for r in pq.roots
		if isnothing(pq.min) || pq.key(r.value) < pq.key(min_element(pq))
			pq.min = r
		end
	end

	return node.value
end

function empty_pq(pq::BinomialHeapPQ)
	return empty_list(pq.roots)
end

function delete(e::Any, pq::BinomialHeapPQ)
	node = lookup(pq.key(e), pq.mapping)
	delete(pq.key(e), pq.mapping)
	delete(node, pq)
end
function delete(node::BinomialHeapNode, pq::BinomialHeapPQ)
	if min_node(pq) != node
		decrease_key(node, pq, typemax(Int))
	end
	if min_node(pq) == node
		delete_min(pq)
	end

	nothing
end

function decrease_key(e::Any, pq::BinomialHeapPQ, delta::Int)
	node = lookup(pq.key(e), pq.mapping)
	decrease_key(node, pq, delta)
end
function decrease_key(node::BinomialHeapNode, pq::BinomialHeapPQ, delta::Int)
	pq.set_key(node.value, pq.key(node.value) - delta)

	sift_up(node, pq)

	nothing
end

function merge(pq::BinomialHeapPQ, pq_::BinomialHeapPQ)
	merge(pq.mapping, pq_.mapping)
	concat(pq.roots, pq_.roots)

	if !empty_pq(pq_) && (isnothing(min_node(pq)) || pq.key(min_element(pq)) > pq_.key(min_element(pq_)))
		pq.min = pq_.min
	end

	trees = trees_same_rank(pq)
	while !isnothing(trees)
		t, t_ = trees
		if pq.key(t.value) < pq.key(t_.value)
			push!(t.children, t_)
			t_.parent = t
			delete(t_, pq.roots)
		else
			push!(t_.children, t)
			t.parent = t_
			delete(t, pq.roots)
		end

		trees = trees_same_rank(pq)
	end

	nothing
end

# Funktionen intern 
function sift_up(node::BinomialHeapNode, pq::BinomialHeapPQ)
	while !isnothing(parent(node)) && pq.key(parent(node).value) > pq.key(node.value)
		# swap parent and e while keeping their parents and children valid

		old_parent = parent(node)
		old_grandparent = parent(old_parent)
		
		# remove e from old_parent's children and add old_parent
		for (index, node_) in enumerate(children(old_parent))
			if node_ == node
				deleteat!(children(old_parent), index)
			end 
		end
		push!(children(old_parent), old_parent)

		if !isnothing(old_grandparent) # not near root list?
			# remove parent from gand parent's children and add me
			for (index, n_) in enumerate(children(old_grandparent))
				if n_ == old_parent
					deleteat!(children(old_grandparent), index)
				end 
			end
			push!(children(old_grandparent), node)
		end

		old_parents_children = children(old_parent)

		old_parent.children = children(node)
		old_parent.parent = node

		node.children = old_parents_children
		node.parent = old_grandparent

		# move e to root list?
		if isnothing(parent(node))
			delete(old_parent, pq.roots)
			insert(node, pq.roots)

			if pq.key(min_element(pq)) > pq.key(node.value)
				pq.min = node
			end
		end
	end

	nothing
end

function trees_same_rank(pq::BinomialHeapPQ)
	# Dies ist nicht optimal, mit einer sortierten Liste koennte man es weiter 
	# beschleunigen.
	for r in pq.roots
		for r_ in pq.roots
			if r != r_ && rank(r) == rank(r_)
				return r, r_
			end 
		end
	end 
	return nothing
end

# Ausgabe in Baumdarstellung durch Julia
function Base.show(io::IO, pq::BinomialHeapPQ)
	print(io, "BinomialHeapPQ mit $(length(pq.roots)) Binomial-Baeumen ")
	println(io, "und $(sum(map(r->2^rank(r), pq.roots))) Elementen")
	println(io, "  min: $(min_element(pq))")
	println(io, "  roots: ")

	l = [map(r->r, pq.roots)]
	while !isempty(l)
		c = l[end]
		if isempty(c)
			pop!(l)
		else
			e = pop!(c)
			pre = "    "
			for i = 1:length(l)-1
				pre = pre * (isempty(l[i]) ? "   " : "|  ")
			end
			println(io, string(pre, "|- ", e))
			if !isempty(children(e))
				push!(l, copy(children(e)))
			end
		end
	end
	#println(io, "key(Element) => Node Mapping: ", pq.mapping)
end
function Base.show(io::IO, node::BinomialHeapNode)
	print(io, node.value)
end

# Beispielaufrufe
# 	Beispieldatentyp
mutable struct Process 
	processID :: Int 
	priority :: Int
end
# 	Ausgabe 
function Base.show(io::IO, e::Process)
	print(io, "$(e.processID) (key=$(e.priority))")
end
# 	Erstellung von "Prozessen" nur durch Angaben der Prioritaet(en)
function Process(p::Int)
	return Process(p, p)
end
function Processes(ps...)
	es = Array{Process}(undef, length(ps))
	for (i,p) in enumerate(ps)
		es[i] = Process(p)
	end
	return es
end
# 	Funktion zum Holen und Setzen der Prioritaet
function get_process_key(e) 
	return e.priority
end
function set_process_key(e, priority) 
	e.priority = priority
end

# Aufrufe
my_es = Processes(9, 11, 3, 1, 7, 5)
my_pq = build(my_es, get_process_key, set_process_key)
println(my_pq)
#	weitere PQ erstellen, fuer merge
my_other_es = Processes(10, 2, 4, 8, 6, 12)
my_other_pq = build(my_other_es, get_process_key, set_process_key)
println(my_other_pq)
#	Merge der beiden PQs
merge(my_pq, my_other_pq)
println(my_pq)
# 	Minima loeschen
println("Minimum loeschen: ", delete_min(my_pq))
println(my_pq)
println("Minimum loeschen: ", delete_min(my_pq))
println(my_pq)
#	Andere loeschen
println("Loesche: ", my_other_es[1])
delete(my_other_es[1], my_pq) 
println(my_pq)
println("Loesche: ", my_other_es[4])
delete(my_other_es[4], my_pq)
println(my_pq)
# 	Wieder einfuegen
my_e = Process(10)
println("Fuege ein: ", my_e)
insert(my_e, my_pq) 
println(my_pq)
#	Key verringern
println("Verringere Key von $(my_es[2]) um 9:")
decrease_key(my_es[2], my_pq, 9)
println(my_pq)