#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 Malte Luttermann
 Henrik Strunck
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Prioritaetswarteschlange mittels Fibonacci-Heap
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Datentypen nutzen!

	Foliensatz: 04-Prioritaetswarteschlangen
	Ab Folie 41
"""

# Linked List analog zu Foliensatz 03, Folie 54 laden
include("./../Utils/linked-list.jl")
include("./../Utils/key-mapping.jl")

# Datentypen
mutable struct FibonacciHeapNode
	value :: Any
	is_marked :: Bool
	parent :: Union{FibonacciHeapNode, Nothing}
	left_sibling :: Union{FibonacciHeapNode, Nothing}
	right_sibling :: Union{FibonacciHeapNode, Nothing}
	first_child :: Union{FibonacciHeapNode, Nothing}
	last_child :: Union{FibonacciHeapNode, Nothing}
end

mutable struct FibonacciHeapPQ
	min :: Union{FibonacciHeapNode, Nothing}
	roots :: LinkedList
	key :: Function 
	set_key :: Function
	mapping :: KeyMapping # Array[key(elment)] = PQ node
end

# Funktionen zum Zugriff auf Attribute
function rank(node::FibonacciHeapNode)
	cntr = 0
	c = first_child(node)
	while !isnothing(c)
		cntr = cntr + 1
		c = right_sibling(c)
	end
	return cntr
end

function parent(node::FibonacciHeapNode)
	return node.parent
end

function first_child(node::FibonacciHeapNode)
	return node.first_child
end

function right_sibling(node::FibonacciHeapNode)
	return node.right_sibling
end

function left_sibling(node::FibonacciHeapNode)
	return node.left_sibling
end

function is_marked(node::FibonacciHeapNode)
	return node.is_marked
end

# Funktionen extern
function build(elements, key::Function, set_key::Function)
	mapping = KeyMapping()
	internal_set_key = function (element, key_value) 
		# Aktualisiert den Key sowohl im Mapping als auch
		# ueber set_key() im Element selbst
		update_key(key(element), key_value, mapping)
		set_key(element, key_value)
	end

	pq = FibonacciHeapPQ(
		nothing, make_list(),
		key, internal_set_key,
		mapping
	)
	for e in elements
		insert(e, pq)
	end
	return pq
end

function insert(e::Any, pq::FibonacciHeapPQ)
	node = FibonacciHeapNode(
		e, false, nothing,
		nothing, nothing,
		nothing, nothing
	)
	insert(pq.key(e), node, pq.mapping)
	insert(node, pq)
end
function insert(node::FibonacciHeapNode, pq::FibonacciHeapPQ)
	pq.roots = cons(node, pq.roots)

	if empty_pq(pq) || isnothing(min_node(pq)) || pq.key(min_element(pq)) > pq.key(node.value)
		pq.min = node
	end

	nothing
end

function min_element(pq::FibonacciHeapPQ)
	return min_node(pq).value
end
function min_node(pq::FibonacciHeapPQ)
	return pq.min
end

function delete_min(pq::FibonacciHeapPQ)
	node = min_node(pq)
	delete(pq.key(node.value), pq.mapping)

	delete(node, pq.roots)
	add_children_to_roots(node, pq)
	# merge zwei Baeume mit Rang i zu Baum mit Rang i+1,
	# solange es entsprechende Baeume gibt (wie bei zwei Binomial-Baeumen)
	full_merge(pq)
	# aktualisiere den min-Pointer
	update_min(pq)

	return node.value
end

function empty_pq(pq::FibonacciHeapPQ)
	return empty_list(pq.roots)
end

function delete(e::Any, pq::FibonacciHeapPQ)
	node = lookup(pq.key(e), pq.mapping)
	delete(pq.key(e), pq.mapping)
	delete(node, pq)
end
function delete(node::FibonacciHeapNode, pq::FibonacciHeapPQ) # lazy delete
	if min_node(pq) == node
		delete_min(pq) # node ist min-Wurzel
	else # haenge node aus, fuege Kinder von node in Wurzelliste ein (entmarkiere entsprechend)
		unhook(node, pq)
		add_children_to_roots(node, pq)
	end
	while !isnothing(parent(node)) && !isnothing(parent(parent(node)))
		node = parent(node)
		if !is_marked(node)
			node.is_marked = true
			return
		else # markiert, also schon Kind weg
			# haenge node samt Unterbaum in Wurzelliste
			unhook(node, pq)
			node.is_marked = false # Wurzeln benoetigen kein Mark
			node.parent = nothing
			insert(node, pq.roots)
		end
	end

	nothing
end

function decrease_key(e::Any, pq::FibonacciHeapPQ, delta::Int)
	node = lookup(pq.key(e), pq.mapping)
	decrease_key(node, pq, delta)
end
function decrease_key(node::FibonacciHeapNode, pq::FibonacciHeapPQ, delta::Int)
	node.is_marked = false
	unhook(node, pq)
	insert(node, pq.roots)
	pq.set_key(node.value, pq.key(node.value) - delta)
	if pq.key(min_element(pq)) > pq.key(node.value)
		pq.min = node
	end

	node_orig = node
	while !isnothing(parent(node)) && !isnothing(parent(parent(node)))
		node = parent(node)
		if !is_marked(node)
			node.is_marked = true
			break
		else # markiert, also schon Kind weg
			# haenge node samt Unterbaum in Wurzelliste
			unhook(node, pq)
			node.is_marked = false # Wurzeln benoetigen kein Mark
			node.parent = nothing
			insert(node, pq.roots)
		end
	end
	node_orig.parent = nothing # Urspruengliches node hat keinen Vater mehr

	nothing
end

function merge(pq::FibonacciHeapPQ, pq_::FibonacciHeapPQ) # lazy merge
	merge(pq.mapping, pq_.mapping)
	concat(pq.roots, pq_.roots)

	if !empty_pq(pq_) && (isnothing(min_node(pq)) || pq.key(min_element(pq)) > pq.key(min_element(pq_)))
		pq.min = pq_.min
	end

	nothing
end

# Funktionen intern 
function unhook(node::FibonacciHeapNode, pq::FibonacciHeapPQ)
	# Haenge Element e samt Unterbaeumen aus pq aus
	if isnothing(parent(node))
		delete(node, pq.roots) # node ist Wurzel
	else
		ls = left_sibling(node)
		rs = right_sibling(node)
		node.left_sibling = nothing
		node.right_sibling = nothing
		if !isnothing(ls)
			ls.right_sibling = rs
		else # node ist erstes Kind
			parent(node).first_child = rs
		end
		if !isnothing(rs)
			rs.left_sibling = ls
		else # node ist letztes Kind
			parent(node).last_child = ls
		end
	end

	nothing
end

function add_children_to_roots(node::FibonacciHeapNode, pq::FibonacciHeapPQ)
	# Haenge alle Unterbaeume von node in die Wurzelliste
	c = first_child(node)
	while !isnothing(c)
		c.parent = nothing
		c.is_marked = false
		insert(c, pq.roots)
		c = right_sibling(c)
	end
	node.first_child = nothing
	node.last_child = nothing
	
	nothing
end

function trees_same_rank(pq::FibonacciHeapPQ)
	# Dies ist nicht optimal, mit einer sortierten Liste koennte man es weiter 
	# beschleunigen.
	for r in pq.roots
		for r_ in pq.roots
			if r != r_ && rank(r) == rank(r_)
				return r, r_
			end 
		end
	end 
	nothing
end

function full_merge(pq::FibonacciHeapPQ) # merge while avoiding trees of the same rank 
	trees = trees_same_rank(pq)
	while !isnothing(trees)
		t, t_ = trees
		if pq.key(t.value) < pq.key(t_.value)
			fc = first_child(t)
			t.first_child = t_
			t_.parent = t
			t_.right_sibling = fc
			if !isnothing(fc)
				fc.left_sibling = t_
			else
				t.last_child = t_
			end
			delete(t_, pq.roots)
		else
			fc = first_child(t_)
			t_.first_child = t
			t.parent = t_
			t.right_sibling = fc
			if !isnothing(fc)
				fc.left_sibling = t
			else
				t_.last_child = t
			end
			delete(t, pq.roots)
		end
		trees = trees_same_rank(pq)
	end
end

function update_min(pq::FibonacciHeapPQ)
	pq.min = nothing 
	for r in pq.roots
		if isnothing(pq.min) || pq.key(r.value) < pq.key(min_element(pq))
			pq.min = r
		end
	end
end

# Ausgabe in Baumdarstellung durch Julia
function Base.show(io::IO, pq::FibonacciHeapPQ)
	println(io, "FibonacciHeapPQ mit $(length(pq.roots)) Baeumen")
	println(io, "  min: $(min_element(pq))")
	println(io, "  roots: ")

	l = [map(r->r, pq.roots)]
	while !isempty(l)
		c = l[end]
		if isempty(c)
			pop!(l)
		else
			e = pop!(c)
			pre = "    "
			for i = 1:length(l)-1
				pre = pre * (isempty(l[i]) ? "   " : "|  ")
			end
			println(io, string(pre, "|- ", e))
			fc = first_child(e)
			children = []
			while !isnothing(fc)
				push!(children, fc)
				fc = right_sibling(fc)
			end
			push!(l, children)
		end
	end
	#println(io, "key(Element) => Node Mapping: ", pq.mapping)
end
function Base.show(io::IO, node::FibonacciHeapNode)
	print(io, node.value)
end

# Beispielaufrufe
# 	Beispieldatentyp
mutable struct Process 
	processID :: Int 
	priority :: Int
end
# 	Ausgabe 
function Base.show(io::IO, e::Process)
	print(io, "$(e.processID) (key=$(e.priority))")
end
# 	Erstellung von "Prozessen" nur durch Angaben der Prioritaet(en)
function Process(p::Int)
	return Process(p, p)
end
function Processes(ps...)
	es = Array{Process}(undef, length(ps))
	for (i,p) in enumerate(ps)
		es[i] = Process(p)
	end
	return es
end
# 	Funktion zum Holen und Setzen der Prioritaet
function get_process_key(e) 
	return e.priority
end
function set_process_key(e, priority) 
	e.priority = priority
end

# Aufrufe
my_es = Processes(9, 11, 3, 1, 7, 5)
my_pq = build(my_es, get_process_key, set_process_key)
println(my_pq)
#	Weitere PQ erstellen, fuer merge
my_other_es = Processes(10, 2, 4, 8, 6, 12)
my_other_pq = build(my_other_es, get_process_key, set_process_key)
println(my_other_pq)
#	Merge der beiden PQs
merge(my_pq, my_other_pq)
println(my_pq)
#	Minima loeschen
println("Minimum loeschen: ", delete_min(my_pq))
println(my_pq)
println("Minimum loeschen: ", delete_min(my_pq))
println(my_pq)
#	Andere loeschen
println("Loesche: ", my_other_es[5])
delete(my_other_es[5], my_pq) 
println(my_pq)
println("Loesche: ", my_other_es[4])
delete(my_other_es[4], my_pq)
println(my_pq)
#	Wieder einfuegen
my_e = Process(15)
println("Fuege ein: ", my_e)
insert(my_e, my_pq) 
println(my_pq)
#	Key verringern
println("Verringere Key von $(my_es[2]) um 9:")
decrease_key(my_es[2], my_pq, 9)
println(my_pq)
#	Minimum loeschen
println("Minimum loeschen: ", delete_min(my_pq))
println(my_pq)
#	Weiteren Key verringern
println("Verringere Key von $(my_es[1]) um 1:")
decrease_key(my_es[1], my_pq, 1)
println(my_pq)
