#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Henrik Strunck
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Binary Search Tree

		Implementierung eines externen binaeren Suchbaumes als Navigationsstruktur,
		die Iterationsstruktur entspricht einer doppelt verketteten zirkulaeren Liste.

	Foliensatz: 06-Mengen
	Ca. Folie(n): 35-59
"""

# Doppelt verkettete Circular Linked List laden
include("./../Utils/circular-linked-list.jl")

# Datentypen
mutable struct TreeNode
	key :: Int
	left :: Union{TreeNode, CircularListElement, Nothing}
	right :: Union{TreeNode, CircularListElement, Nothing}
	parent :: Union{TreeNode, Nothing}
end

mutable struct BinarySearchTree
	root :: Union{TreeNode, Nothing}
end

mutable struct SearchStructure
	navigation_repr :: Union{BinarySearchTree, Nothing}
	internal_repr :: Union{CircularLinkedList, Nothing}
	key :: Function
end

# Funktionen extern
function make_search_structure(key::Function=identity) :: SearchStructure
	return SearchStructure(BinarySearchTree(nothing), make_circular_list(), key)
end

function search(k::Int, s::SearchStructure) :: Union{Any, Nothing}
	return search(k, s.navigation_repr.root)
end
function search(k::Int, v::Union{TreeNode, Nothing}) :: Union{Any, TreeNode, Nothing}
	if !isnothing(v)
		if k <= v.key
			return search(k, v.left)
		else
			return search(k, v.right)
		end
	end
	return nothing
end
function search(k::Int, le::CircularListElement) :: Any
	return le.value
end

function insert(e::Any, s::SearchStructure) :: Nothing
	if isnothing(s.navigation_repr.root) # tree is empty
		new_element = insert_first(e, s.internal_repr)
		s.navigation_repr.root = TreeNode(s.key(e), new_element, nothing, nothing)
	else
		p, le = search_internal(s.key(e), s.navigation_repr.root, s.navigation_repr.root)
		if isnothing(le) # no greater list element
			new_element = insert_last(e, s.internal_repr)
			p.right = TreeNode(s.key(e), new_element, nothing, p)
		elseif s.key(e) != s.key(le.value)
			# insert e before le and add tree node
			new_element = insert_before(e, le, s.internal_repr)
			if s.key(e) < p.key
				p.left = TreeNode(s.key(e), new_element, le, p)
			else
				p.right = TreeNode(s.key(e), new_element, le, p)
			end
		else
			# replace existing object with new one but same key	(no duplicates)
			le.value = e
		end
	end
	nothing
end

function delete(k::Int, s::SearchStructure) :: Nothing
	if !isnothing(s.navigation_repr.root)
		p, le = search_internal(k, s.navigation_repr.root, s.navigation_repr.root)
		if !isnothing(le) && s.key(le.value) == k
			if isnothing(p.parent) # parent of element is root
				s.navigation_repr.root = nothing
			elseif s.key(le.value) <= p.key # element is left child and parent no root
				if p.key <= p.parent.key # tree node is left child
					p.parent.left = p.right
				else # tree node is right child
					p.parent.right = p.right
				end
			else # element is right child and parent no root
				# search tree node with key to be deleted
				node = p.parent
				while node.key != k
					node = node.parent
				end
				node.key = p.key
				if p.key <= p.parent.key # tree node is left child
					p.parent.left = p.left
				else # tree node is right child
					p.parent.right = p.left
				end				
			end
			delete(le, s.internal_repr)
		end
	end
	nothing
end

function empty_tree(v::Union{TreeNode, Nothing}) :: Bool
	return isnothing(v)
end
function empty_tree(le::CircularListElement) :: Bool
	return true
end

function is_leaf(v::TreeNode) :: Bool
	return typeof(v.left) != TreeNode && typeof(v.right) != TreeNode
end

function left_exists(v::TreeNode) :: Bool
	return !isnothing(v.left)
end

function right_exists(v::TreeNode) :: Bool
	return !isnothing(v.right)
end

function fold(f::Function, s::SearchStructure, init::Any)
	return fold(f, s.navigation_repr.root, init)
end
function fold(f::Function, v::Union{TreeNode, CircularListElement, Nothing}, init::Any)
	if empty_tree(v)
		return init
	end
	if is_leaf(v)
		return f(init, v.key)
	end
	if left_exists(v)
		x = f(fold(f, v.left, init), v.key)
		if right_exists(v)
			return fold(f, v.right, x)
		else
			return x
		end
	else # linker Nachfolger existiert nicht, rechter ja, sonst leaf
		return fold(f, v.right, f(init, v.key))
	end
end

# Funktionen intern
function search_internal(k::Int, p::TreeNode, v::Union{TreeNode, Nothing}) :: Tuple{TreeNode, Union{CircularListElement, Nothing}}
	if !isnothing(v)
		if k <= v.key
			return search_internal(k, v, v.left)
		else
			return search_internal(k, v, v.right)
		end
	end
	return p, nothing
end
function search_internal(k::Int, p::TreeNode, le::CircularListElement) :: Tuple{TreeNode, CircularListElement}
	return p, le
end

function print_tree_node(x, k::Int) :: Int
	print("-> $(k)")
	return 0
end

# Ausgabe in Baumdarstellung durch Julia
function Base.show(io::IO, s::SearchStructure)
	println(io, "<=================== Search Structure: ")
	print(io, s.navigation_repr)
	println(io, s.internal_repr)
	println(io, "===================>")
end
function Base.show(io::IO, bst::BinarySearchTree)
	println(io, "Binary Search Tree:")
	if isnothing(bst.root)
		println("\tEmpty Tree")
	else
		print(io, bst.root)
	end
end
function Base.show(io::IO, v::TreeNode)
	show(io, v, 0, true, false)
end
function Base.show(io::IO, v::TreeNode, indent::Int, is_root::Bool, is_left::Bool)
	if !empty_tree(v.right)
		show(io, v.right, indent+1, false, false)
	end
	node_str = if is_root "-< " elseif is_left "\\- " else "/- " end 	
	println(io, repeat("    ", indent), node_str, v.key)
	if !empty_tree(v.left)
		show(io, v.left, indent+1, false, true)
	end
end

# Beispielaufrufe
my_search_struct = make_search_structure()

println("Fuege 3 ein:")
insert(3, my_search_struct)
println(my_search_struct)

println("Fuege 6 ein:")
insert(6, my_search_struct)
println(my_search_struct)

println("Fuege 4 ein:")
insert(4, my_search_struct)
println(my_search_struct)

println("Fuege 5 ein:")
insert(5, my_search_struct)
println(my_search_struct)

println("Fuege 1 ein:")
insert(1, my_search_struct)
println(my_search_struct)

println("Fuege 2 ein:")
insert(2, my_search_struct)
println(my_search_struct)

println("Suche nach 2: " * string(search(2, my_search_struct)))
println("Suche nach 0: " * string(search(0, my_search_struct)))
println("Suche nach 42: " * string(search(42, my_search_struct)))
println("")

println("Loeschen von 3:")
delete(3, my_search_struct)
println(my_search_struct)

println("Loeschen von 4:")
delete(4, my_search_struct)
println(my_search_struct)

println("Loeschen von 1:")
delete(1, my_search_struct)
println(my_search_struct)
