#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Splay Tree

		Implementierung eines Splay Baumes als internen Suchbaum (d.h. Elemente
		sind direkt integriert in Baum und nicht in extra Liste "unten").

	Foliensatz: 06-Mengen
	Ca. ab Folie(n): 67
"""

mutable struct SplayTreeNode
	value::Any
	parent::Union{SplayTreeNode, Nothing}
	left_c::Union{SplayTreeNode, Nothing}
	right_c::Union{SplayTreeNode, Nothing}
end

function set_left_child(v::SplayTreeNode, c::Union{SplayTreeNode, Nothing})::Nothing
	v.left_c = c
	if !isnothing(c)
		c.parent = v
	end
	nothing
end

function set_right_child(v::SplayTreeNode, c::Union{SplayTreeNode, Nothing})::Nothing
	v.right_c = c
	if !isnothing(c)
		c.parent = v
	end
	nothing
end

mutable struct SplayTree
	root::Union{SplayTreeNode, Nothing}
	key::Function
end

function make_splay_tree(key::Function)::SplayTree
	return SplayTree(nothing, key)
end
function make_splay_tree(elements, key::Function)::SplayTree
	t = make_splay_tree(key)
	for e in elements
		insert(e, t)
	end
	return t
end

function insert(e::Any, t::SplayTree)::Nothing
	if isnothing(t.root) # Baum ist leer, e wird neue Wurzel
		t.root = SplayTreeNode(e, nothing, nothing, nothing)
	else
		v = t.root # Beginne mit Wurzel und suche passende Stelle zum Einfuegen
		while true
			if t.key(e) == t.key(v.value) # Schluessel existiert schon, ersetze Wert
				v.value = e
				return
			elseif t.key(e) < t.key(v.value) # Suche im linken Teilbaum
				if isnothing(v.left_c) # Passende Stelle gefunden, fuege e ein
					set_left_child(v, SplayTreeNode(e, v, nothing, nothing))
					splay(v.left_c, t)
					return
				else # Fahre mit Suche im linken Teilbaum fort
					v = v.left_c
				end
			else # t.key(e) > t.key(v.value), Suche im rechten Teilbaum
				if isnothing(v.right_c) # Passende Stelle gefunden, fuege e ein
					set_right_child(v, SplayTreeNode(e, v, nothing, nothing))
					splay(v.right_c, t)
					return
				else # Fahre mit Suche im rechten Teilbaum fort
					v = v.right_c
				end
			end
		end
	end
	nothing
end

function merge(v_l::Union{SplayTreeNode, Nothing}, v_r::Union{SplayTreeNode, Nothing}) :: SplayTreeNode
	max_left = max(v_l) # finde Maximum im linken Teilbaum

	if !isnothing(max_left.left_c) && max_left != v_l
		set_right_child(max_left.parent, max_left.left_c)
	end
	max_left.parent = nothing # neue Wurzel 

	# setze Kinder und Eltern
	set_right_child(max_left, v_r)
	if max_left != v_l
		set_left_child(max_left, v_l) 
	end

	return max_left
end

function delete(e::Any, t::SplayTree)::Nothing
	if isnothing(search(e, t))
		return # Element nicht gefunden
	end
	# Element ist nun in Wurzel, durch search()

	# merge ohne "Wurzel" (zu loeschenden Knoten)
	t.root = merge(t.root.left_c, t.root.right_c)

	nothing
end

function search(e::Any, t::SplayTree)::Union{Any, Nothing}
	v = t.root
	while !isnothing(v)
		if t.key(e) == t.key(v.value)
			# Schluessel gefunden, wende Splay-Operation an und gib
			# das entsprechende Element zurueck
			splay(v, t)
			return v.value
		elseif t.key(e) < t.key(v.value)
			v = v.left_c # Suche weiter im linken Teilbaum
		else
			v = v.right_c # Suche weiter im rechten Teilbaum
		end
	end
	return nothing
end

function splay(x::SplayTreeNode, t::SplayTree)::Nothing
	while !isnothing(x.parent)
		y = x.parent
		if isnothing(y.parent) # x ist direktes Kind der Wurzel
			x.parent = nothing # x ist neue Wurzel (wird unten gesetzt)
			if y.left_c == x # x ist linkes Kind von y (Fall 1a)
				set_left_child(y, x.right_c)
				set_right_child(x, y)
			elseif y.right_c == x # x ist rechtes Kind von y (Fall 1b)
				set_right_child(y, x.left_c)
				set_left_child(x, y)
			end
		else # x ist kein Kind der Wurzel (Faelle 2a, 2b, 3a, 3b)
			z = y.parent
			if isnothing(z.parent)
				x.parent = nothing
			else
				set_parent_child(z, x, t)
			end
			if z.left_c == y # y ist linkes Kind von z (Faelle 2a, 3a)
				if y.left_c == x # x ist linkes Kind von y (Fall 2a)
					set_left_child(y, x.right_c)
					set_right_child(x, y)
					set_left_child(z, y.right_c)
					set_right_child(y, z)
				elseif y.right_c == x # x ist rechtes Kind von y (Fall 3a)
					set_right_child(y, x.left_c)
					set_left_child(x, y)
					set_left_child(z, x.right_c)
					set_right_child(x, z)
				end
			elseif z.right_c == y # y ist rechtes Kind von z (Faelle 2b, 3b)
				if y.right_c == x # x ist rechtes Kind von y (Fall 2b)
					set_right_child(y, x.left_c)
					set_left_child(x, y)
					set_right_child(z, y.left_c)
					set_left_child(y, z)
				elseif y.left_c == x # x ist linkes Kind von y (Fall 3b)
					set_left_child(y, x.right_c)
					set_right_child(x, y)
					set_right_child(z, x.left_c)
					set_left_child(x, z)
				end
			end
		end
	end
	t.root = x
	nothing
end

# Hilfsfunktionen
function max(v::SplayTreeNode)::SplayTreeNode
	# Ermittelt den maximalen Schluessel im Teilbaum v
	current = v
	while !isnothing(current.right_c)
		current = current.right_c
	end
	return current
end

function set_parent_child(v::SplayTreeNode, new_c::Union{Nothing, SplayTreeNode}, t::SplayTree)::Nothing
	# Setzt das linke Kind des Elternknotens von v, falls v linkes Kind ist
	# bzw. das rechte Kind des Elternknotens von v, falls v rechtes Kind ist
	# auf den neuen Wert `new_c`.
	# Falls es keinen Elternknoten gibt, wird die Wurzel auf `new_c` gesetzt.
	if isnothing(v.parent) # v ist Wurzel
		t.root = new_c
		if !isnothing(new_c)
			t.root.parent = nothing
		end
	elseif v.parent.left_c == v # v ist linkes Kind
		set_left_child(v.parent, new_c)
	else # v ist rechtes Kind
		set_right_child(v.parent, new_c)
	end
	nothing
end

# Ausgabe in Baumdarstellung durch Julia
function Base.show(io::IO, t::SplayTree)
	println(io, "Splay Tree:")
	if isnothing(t.root)
		println("\tEmpty Tree")
	else
		print(io, t.root)
	end
end
function Base.show(io::IO, v::SplayTreeNode)
	show(io, v, 0, true, false)
end
function Base.show(io::IO, v::SplayTreeNode, indent::Int, is_root::Bool, is_left::Bool)
	if !isnothing(v.right_c)
		show(io, v.right_c, indent+1, false, false)
	end
	node_str = if is_root "-< " elseif is_left "\\- " else "/- " end 	
	println(io, repeat("    ", indent), node_str, v.value)
	if !isnothing(v.left_c)
		show(io, v.left_c, indent+1, false, true)
	end
end


# Beispielaufrufe
t = make_splay_tree([10, 1, 14, 28], identity)
println(t)

println("Fuege 5 ein:")
insert(5, t)
println(t)
println("Fuege 12 ein:")
insert(12, t)
println(t)

println("Suche 10:")
println(search(10, t))
println(t)
println("Suche 12:")
println(search(12, t))
println(t)

println("Loesche 28:")
delete(28, t)
println(t)
println("Loesche 10:")
delete(10, t)
println(t)

println("===")
# Weitere Beispielaufrufe mit benutzerdefiniertem Beispieldatentyp
mutable struct Student
	matr_nr::Int
	name::String
end
elements = [
	Student(10, "Max"),
	Student(1, "Alice"),
	Student(14, "Peter"),
	Student(28, "Maria")
]

t_ = make_splay_tree(elements, (e) -> e.matr_nr)
println(t_)

println("Fuege (12, Bob) ein:")
bob = Student(12, "Bob")
insert(bob, t_)
println(t_)

println("Suche (12, Bob):")
println(search(bob, t_))
println(t_)

println("Loesche (12, Bob):")
delete(bob, t_)
println(t_)