#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Red Black Tree

		Implementierung eines Rot-Schwarz Baumes als internen Suchbaum (d.h. Elemente
		sind direkt integriert in Baum und nicht in extra Liste "unten").

	Foliensatz: 06-Mengen
	Ca. ab Folie(n): 94
"""

mutable struct RSTreeNode
	value::Any
	is_red::Bool
	parent::Union{RSTreeNode, Nothing}
	left_c::Union{RSTreeNode, Nothing}
	right_c::Union{RSTreeNode, Nothing}

	RSTreeNode(e) = new(e, true, nothing, nothing, nothing)
end

function set_left_child(v::RSTreeNode, c::Union{RSTreeNode, Nothing}) :: Nothing
	v.left_c = c
	if !isnothing(c)
		c.parent = v
	end
	nothing
end
function set_right_child(v::RSTreeNode, c::Union{RSTreeNode, Nothing}) :: Nothing
	v.right_c = c
	if !isnothing(c)
		c.parent = v
	end
	nothing
end

function make_red(v::Union{RSTreeNode, Nothing}) :: Nothing
	if !isnothing(v)
		v.is_red = true
	end
	nothing
end
function is_red(v::Union{RSTreeNode, Nothing}) :: Bool
	return !isnothing(v) && v.is_red
end

function make_black(v::Union{RSTreeNode, Nothing}) :: Nothing
	if !isnothing(v)
		v.is_red = false
	end
	nothing
end
function is_black(v::Union{RSTreeNode, Nothing}) :: Bool
	# Null-Knoten sind auch schwarz
	return isnothing(v) || !v.is_red
end

function uncle(v::RSTreeNode) :: Union{RSTreeNode, Nothing}
	if !isnothing(v.parent.parent.left_c) && v.parent.parent.left_c != v.parent
		return v.parent.parent.left_c
	elseif !isnothing(v.parent.parent.right_c) && v.parent.parent.right_c != v.parent
		return v.parent.parent.right_c
	end
	return nothing
end

function brother(v::RSTreeNode) :: Union{RSTreeNode, Nothing}
	if !isnothing(v.parent.left_c) && v.parent.left_c != v
		return v.parent.left_c
	elseif !isnothing(v.parent.right_c) && v.parent.right_c != v
		return v.parent.right_c
	end
	return nothing
end

function is_left_child(v::RSTreeNode) :: Bool 
	return v.parent.left_c == v
end
function is_right_child(v::RSTreeNode) :: Bool 
	return v.parent.right_c == v
end

mutable struct RSTree
	root::Union{RSTreeNode, Nothing}
	key::Function
end

function make_rs_tree(key::Function) :: RSTree
	return RSTree(nothing, key)
end
function make_rs_tree(elements, key::Function) :: RSTree
	t = make_rs_tree(key)
	for e in elements
		insert(e, t)
	end
	return t
end

function insert(e::Any, t::RSTree) :: Nothing
	if isnothing(t.root) # Baum ist leer, e wird neue Wurzel
		t.root = RSTreeNode(e)
		rotate_insert(t.root, t)
	else
		v = t.root # Beginne mit Wurzel und suche passende Stelle zum Einfuegen
		while true
			if t.key(e) == t.key(v.value) # Schluessel existiert schon, ersetze Wert
				v.value = e
				return
			elseif t.key(e) < t.key(v.value) # Suche im linken Teilbaum
				if isnothing(v.left_c) # Passende Stelle gefunden, fuege e ein
					set_left_child(v, RSTreeNode(e))
					rotate_insert(v.left_c, t)
					return
				else # Fahre mit Suche im linken Teilbaum fort
					v = v.left_c
				end
			else # t.key(e) > t.key(v.value), Suche im rechten Teilbaum
				if isnothing(v.right_c) # Passende Stelle gefunden, fuege e ein
					set_right_child(v, RSTreeNode(e))
					rotate_insert(v.right_c, t)
					return
				else # Fahre mit Suche im rechten Teilbaum fort
					v = v.right_c
				end
			end
		end
	end
	nothing
end

function rotate_insert(v::RSTreeNode, t::RSTree) :: Nothing
	# Fuehre die Rotationen und Umfaerbungen durch, nachdem 
	# neuer Knoten v wie in BSB eingefuegt wurde
	#	analog zu PDF: "Zusammenfassung: ROT-SCHWARZ-Baeume"
	if isnothing(v.parent) # 1. Wurzel
		v.is_red = false 
	elseif !v.parent.is_red # 2. Vater schwarz 
		v.is_red = true 
	else # 3. Vater von v rot
		if is_black(uncle(v)) # (a) Fall 1
			p = v.parent # Vater (Parent)
			g = p.parent # Grossvater (Grandparent)

			gg = g.parent # Urgrossvater (Great grandparent)
			if !isnothing(gg) 
				# wir muessen spaeter die Teilbaumwurzel
				# richtig als Kind des Urgrossvaters festlegen
				g_was_left = is_left_child(g) # Grossvater linkes Kind
			end

			if is_left_child(p) && is_left_child(v) # i. (einfach)
				make_black(p)
				make_red(g)
				# Rechts, Rechts-Rotation (Folie 122, 123)
				set_left_child(g, p.right_c)
				set_right_child(p, g)
				subtree_root = p
			elseif is_right_child(p) && is_right_child(v) # i. (einfach)
				make_black(p)
				make_red(g)
				# Links, Links-Rotation (Folie 127)
				set_right_child(g, p.left_c)
				set_left_child(p, g)
				subtree_root = p
			elseif is_left_child(p) && is_right_child(v) # ii. (doppelt)
				make_black(v)
				make_red(g)
				# Links, Rechts-Rotation
				set_right_child(p, v.left_c)
				set_left_child(v, p)
				set_left_child(g, v.right_c)
				set_right_child(v, g)
				subtree_root = v
			elseif is_right_child(p) && is_left_child(v) # ii. (doppelt)
				make_black(v)
				make_red(g)
				# Rechts, Links-Rotation
				set_left_child(p, v.right_c)
				set_right_child(v, p)
				set_right_child(g, v.left_c)
				set_left_child(v, g)
				subtree_root = v
			end

			if isnothing(gg) # kein Urgrossvater, also war Grossvater Wurzel
				t.root = subtree_root 
				subtree_root.parent = nothing
			else
				# Teilbaumwurzel als richtiges Kind des des Urgrossvaters
				# festlegen
				if g_was_left
					set_left_child(gg, subtree_root)
				else
					set_right_child(gg, subtree_root)
				end
			end
		else # (b) Fall 2
			make_black(v.parent)
			make_black(uncle(v))
			make_red(v.parent.parent)
			rotate_insert(v.parent.parent, t::RSTree)
		end
	end

	nothing
end

function delete(e::Any, t::RSTree) :: Nothing
	v = t.root
	while !isnothing(v)
		if t.key(e) == t.key(v.value) # Schluessel gefunden
			# Fallunterscheidung anhand der Kinder von v:
			if isnothing(v.left_c) && isnothing(v.right_c)
				# v hat keine Kinder: Entferne v direkt aus dem Baum
				# indem dem Elternknoten mitgeteilt wird, dass sein
				# Kind nun `nothing` ist
				rotate_delete(v, nothing, t)
			elseif isnothing(v.left_c) # v hat nur rechtes Kind
				# Neues Kind des Elternknotens ist nun rechtes Kind von v
				rotate_delete(v, v.right_c, t)
			elseif isnothing(v.right_c) # v hat nur linkes Kind
				# Neues Kind des Elternknotens ist nun linkes Kind von v
				rotate_delete(v, v.left_c, t)
			else # v hat sowohl linkes Kind als auch rechtes Kind
				# Bestimme Maximum im linken Teilbaum, uebernehme Wert
				# des Maximums in v und loesche das Maximum aus dem linken
				# Teilbaum (hat garantiert hoechstens ein Kind)
				max_left = max(v.left_c) # Finde Maximum im linken Teilbaum
				v.value = max_left.value # Uebernehme Wert des Maximums
				rotate_delete(max_left, max_left.left_c, t)
			end
			return # Schluessel wurde gefunden und geloescht, fertig
		elseif t.key(e) < t.key(v.value) # Suche weiter im linken Teilbaum
			v = v.left_c
		else # t.key(e) > t.key(v.value), Suche weiter im rechten Teilbaum
			v = v.right_c
		end
	end
	nothing
end

function rotate_delete(v::Union{RSTreeNode, Nothing}, r::Union{RSTreeNode, Nothing}, t::RSTree)
	# Fuehre die Rotationen und Umfaerbungen durch, und 
	# loesche Knoten v anschliessen (Knoten v hat hier max. 1 Kind naemlich r)
	#	analog zu PDF: "Zusammenfassung: ROT-SCHWARZ-Baeume"

	if is_red(v) # 1. v Rot
		set_parent_child(v, r, t)
	elseif is_black(v)  # 2. v Schwarz
		set_parent_child(v, r, t) 

		while true
			if t.root == r # (a) r ist zur Wurzel geworden
				if is_red(r)
					make_black(r)
				end
				return

			else # (b) v nicht Wurzel 
				if is_red(r) # (i) Rot
					make_black(r)
					return

				else  # (ii) Schwarz
					y = brother(v) # == brother(r) (aber r koennte nothing sein)
					x = v.parent # == r.parent (aber r koennte nothing sein)
					if is_black(y) && !isnothing(y) && (is_red(y.left_c) || is_red(y.right_c))
						# A. Fall 1
						z = if is_red(y.left_c) y.left_c else y.right_c end

						if is_left_child(y) && is_left_child(z) # - (einfach)
							if is_red(x) make_red(y) else make_black(y) end 
							make_black(x)
							make_black(z)
							# Rechtsrotation
							set_parent_child(x, y, t)
							set_left_child(x, y.right_c)
							set_right_child(y, x)

						elseif is_right_child(y) && is_right_child(z) # - (einfach)
							if is_red(x) make_red(y) else make_black(y) end 
							make_black(x)
							make_black(z)
							# Linksrotation
							set_parent_child(x, y, t)
							set_right_child(x, y.left_c)
							set_left_child(y, x)

						elseif is_left_child(y) && is_right_child(z) # - (doppelt)
							make_black(z)
							if is_red(x) make_red(y) else make_black(y) end 
							# Links, Rechtsrotation
							set_parent_child(x, z, t)
							set_right_child(y, z.left_c)
							set_left_child(x, z.right_c)
							set_right_child(z, x)
							set_left_child(z, y)

						elseif is_right_child(y) && is_left_child(z) # - (doppelt)
							make_black(z)
							if is_red(x) make_red(y) else make_black(y) end 
							# Rechts, Linksrotation
							set_parent_child(x, z, t)
							set_left_child(y, z.right_c)
							set_right_child(x, z.left_c)
							set_left_child(z, x)
							set_right_child(z, y)

						end
						return

					elseif is_black(y) && !isnothing(y) && is_black(y.left_c) && is_black(y.right_c)
						# B. Fall 2
						make_red(y)
						r = y.parent
						v = y.parent # auch fuer "altes Kind"

					elseif is_red(y)
						# C. Fall 3
						if is_right_child(y) # anders herum, um Fall r == nothing moeglich zu machen, 
							#	denn es gilt is_left_child(r) <=> is_right_child(y)
							set_parent_child(x, y, t)
							set_right_child(x, y.left_c)
							set_left_child(y, x)
						elseif is_left_child(y)
							set_parent_child(x, y, t)
							set_left_child(x, y.right_c)
							set_right_child(y, x)
						end
						make_red(x)
						if !isnothing(x.parent) 
							make_black(x.parent)
						end
					else 
						return
					end
				end
			end
		end
	end
end

function search(e::Any, t::RSTree) :: Union{Any, Nothing}
	v = t.root
	while !isnothing(v)
		if t.key(e) == t.key(v.value)
			# Schluessel gefunden, das entsprechende Element zurueck
			return v.value
		elseif t.key(e) < t.key(v.value)
			v = v.left_c # Suche weiter im linken Teilbaum
		else
			v = v.right_c # Suche weiter im rechten Teilbaum
		end
	end
	return nothing
end

# Hilfsfunktionen
function max(v::RSTreeNode) :: RSTreeNode
	# Ermittelt den maximalen Schluessel im Teilbaum v
	current = v
	while !isnothing(current.right_c)
		current = current.right_c
	end
	return current
end

function set_parent_child(v::RSTreeNode, new_c::Union{Nothing, RSTreeNode}, t::RSTree)::Nothing
	# Setzt das linke Kind des Elternknotens von v, falls v linkes Kind ist
	# bzw. das rechte Kind des Elternknotens von v, falls v rechtes Kind ist
	# auf den neuen Wert `new_c`.
	# Falls es keinen Elternknoten gibt, wird die Wurzel auf `new_c` gesetzt.
	if isnothing(v.parent) # v ist Wurzel
		t.root = new_c
		if !isnothing(new_c)
			t.root.parent = nothing
		end
	elseif v.parent.left_c == v # v ist linkes Kind
		set_left_child(v.parent, new_c)
	else # v ist rechtes Kind
		set_right_child(v.parent, new_c)
	end
	nothing
end

# Ausgabe in Baumdarstellung durch Julia
function Base.show(io::IO, t::RSTree)
	println(io, "Red Black Tree:")
	if isnothing(t.root)
		println("\tEmpty Tree")
	else
		show(io, t.root, 0, true, false)
	end
end
function Base.show(io::IO, v::RSTreeNode, indent::Int, is_root::Bool, is_left::Bool)
	if !isnothing(v.right_c)
		show(io, v.right_c, indent+1, false, false)
	end
	print(io, repeat("    ", indent), if is_root "-< " elseif is_left "\\- " else "/- " end)
	printstyled(io, v.value, color=Symbol(if v.is_red "red" else "black" end), bold=true)
	println()
	if !isnothing(v.left_c)
		show(io, v.left_c, indent+1, false, true)
	end
end

# Beispielaufrufe
t = make_rs_tree([9, 1, 14, 11, 28], identity)
println(t)

for e in [5, 22, 21, 30, 12, 10, 0]
	println("Fuege $e ein:")
	insert(e, t)
	println(t)
end

for e in [100, -1, 3, 12, 30]
	println("Suche $e: ", search(e, t))
end

for e in [21, 30, 5, 22, 12, 10, 0, 11, 1, 14, 9, 28]
	println("Loesche $e:")
	delete(e, t)
	println(t)
end