#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Henrik Strunck
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	AVL Tree

		Implementierung eines AVL-Baumes als internen Suchbaum (d.h. Elemente
		sind direkt integriert in Baum und nicht in extra Liste "unten").

	Foliensatz: 06-Mengen
	Ca. ab Folie(n): 129
"""

mutable struct AVLTreeNode
	value :: Any
	height :: Int
	parent :: Union{AVLTreeNode, Nothing}
	left_c :: Union{AVLTreeNode, Nothing}
	right_c :: Union{AVLTreeNode, Nothing}

	AVLTreeNode(e) = new(e, 1, nothing, nothing, nothing)
	AVLTreeNode(e, p) = new(e, 1, p, nothing, nothing)
end

mutable struct AVLTree
	root :: Union{AVLTreeNode, Nothing}
	key :: Function
end

function make_avl_tree(key::Function=identity) :: AVLTree
	return AVLTree(nothing, key)
end
function make_avl_tree(elements, key::Function=identity) :: AVLTree
	t = make_avl_tree(key)
	for e in elements
		insert(e, t)
	end
	return t
end

function set_left_child(v::AVLTreeNode, c::Union{AVLTreeNode, Nothing}) :: Nothing
	v.left_c = c
	if !isnothing(c)
		c.parent = v
	end
	nothing
end
function set_right_child(v::AVLTreeNode, c::Union{AVLTreeNode, Nothing}) :: Nothing
	v.right_c = c
	if !isnothing(c)
		c.parent = v
	end
	nothing
end

function rotate_left(v::AVLTreeNode, t::AVLTree) :: AVLTreeNode
	right_c = v.right_c
	set_parent_child(v, right_c, t)
	set_right_child(v, right_c.left_c)
	set_left_child(right_c, v)
	update_height(v)
	update_height(right_c)
	return right_c
end

function rotate_right(v::AVLTreeNode, t::AVLTree) :: AVLTreeNode
	left_c = v.left_c
	set_parent_child(v, left_c, t)
	set_left_child(v, left_c.right_c)
	set_right_child(left_c, v)
	update_height(v)
	update_height(left_c)
	return left_c
end

function balance_factor(v::AVLTreeNode) :: Int
	height_left = 0
	height_right = 0
	if !isnothing(v.left_c)
		height_left = v.left_c.height
	end
	if !isnothing(v.right_c)
		height_right = v.right_c.height
	end
	return height_right - height_left
end

function update_height_tree(v::AVLTreeNode) :: Nothing
	# Die Funktion aktualisiert die Hoehe auf dem Pfad
	# von v bis zur Wurzel
	while !isnothing(v)
		update_height(v)
		v = v.parent
	end
	nothing
end
function update_height(v::AVLTreeNode) :: Nothing
	# Die Funktion aktualisiert die Hoehe von v
	height_left = 0
	height_right = 0
	if !isnothing(v.left_c)
		height_left = v.left_c.height
	end
	if !isnothing(v.right_c)
		height_right = v.right_c.height
	end
	v.height = Base.max(height_left, height_right) + 1
	nothing
end

function rebalance_after_insertion(v::AVLTreeNode, t::AVLTree) :: Nothing
	# Die Funktion rebalanciert, sofern notwendig, den Baum
	# auf dem Pfad von v bis zur Wurzel an der tiefsten Stelle und
	# aktualisiert auch die Hoehe
	while !isnothing(v)
		subtree_root = rebalance(v, t)
		# Eine Einfach- oder Doppelrotation ist ausreichend, um den
		# Baum wieder auszugleichen
		if subtree_root != v && !isnothing(subtree_root.parent)
			update_height_tree(subtree_root.parent)
			break
		end
		v = v.parent
	end
end

function rebalance_after_deletion(v::AVLTreeNode, t::AVLTree) :: Nothing
	# Die Funktion rebalanciert iterativ, sofern notwendig, den Baum
	# auf dem Pfad von v bis zur Wurzel an notwendigen Stellen
	# und aktualisiert dabei jeweils auch die Hoehe
	while !isnothing(v)
		subtree_root = rebalance(v, t)
		# Aktualisiere Hoehe, falls in dieser Iteration keine Rotation
		# notwendig war, aber ggf. tiefer im Baum bereits rotiert wurde
		if subtree_root == v
			update_height(subtree_root)
		end
		v = v.parent
	end
end

function rebalance(v::AVLTreeNode, t::AVLTree) :: AVLTreeNode
	# Die Funktion rebalanciert, sofern notwendig,
	# den Teilbaum mit Wurzel v und setzt dafuer Rotationen ein.
	# Es wird die ggf. neue Wurzel des Teilbaums zurückgegeben
	bf = balance_factor(v)
	subtree_root = v
	if bf < -1 # linkslastig
		if balance_factor(v.left_c) < 0
			subtree_root = rotate_right(v, t) # rechts rotieren
		else
			rotate_left(v.left_c, t) # links-rechts rotieren
			subtree_root = rotate_right(v, t)
		end
	elseif bf > 1 # rechtslastig
		if balance_factor(v.right_c) > 0
			subtree_root = rotate_left(v, t) # links rotieren
		else
			rotate_right(v.right_c, t) # rechts-links rotieren
			subtree_root = rotate_left(v, t)
		end
	end
	return subtree_root
end

function insert(e::Any, t::AVLTree) :: Nothing
	if isnothing(t.root) # Baum ist leer, e wird neue Wurzel
		t.root = AVLTreeNode(e)
	else
		v = t.root # Beginne mit Wurzel und suche passende Stelle zum Einfuegen
		while true
			if t.key(e) == t.key(v.value) # Schluessel existiert schon, ersetze Wert
				v.value = e
				return
			elseif t.key(e) < t.key(v.value) # Suche im linken Teilbaum
				if isnothing(v.left_c) # Passende Stelle gefunden, fuege e ein
					v.left_c = AVLTreeNode(e, v)
					update_height_tree(v) # Hoehe auf dem Pfad zur Wurzel aktualisieren
					if !isnothing(v.parent) # Baum kann nur ab grandparent unbalanciert sein
						if @isdefined(SHOW_BEFORE_REBALANCE) && SHOW_BEFORE_REBALANCE
							println("Before rebalance:")
							println(t)
						end
						subtree_root = v.parent
						rebalance_after_insertion(subtree_root, t)
					end
					return
				else # Fahre mit Suche im linken Teilbaum fort
					v = v.left_c
				end
			else # t.key(e) > t.key(v.value), Suche im rechten Teilbaum
				if isnothing(v.right_c) # Passende Stelle gefunden, fuege e ein
					v.right_c = AVLTreeNode(e, v)
					update_height_tree(v) # Hoehe auf dem Pfad zur Wurzel aktualisieren
					if !isnothing(v.parent) # Baum kann nur ab grandparent unbalanciert sein
						if @isdefined(SHOW_BEFORE_REBALANCE) && SHOW_BEFORE_REBALANCE
							println("Before rebalance:")
							println(t)
						end
						subtree_root = v.parent
						rebalance_after_insertion(subtree_root, t)
					end
					return
				else # Fahre mit Suche im rechten Teilbaum fort
					v = v.right_c
				end
			end
		end
	end
	nothing
end

function delete(e::Any, t::AVLTree) :: Nothing
	v = t.root
	while !isnothing(v)
		if t.key(e) == t.key(v.value) # Schluessel gefunden
			# Fallunterscheidung anhand der Kinder von v:
			if isnothing(v.left_c) && isnothing(v.right_c)
				# v hat keine Kinder: Entferne v direkt aus dem Baum,
				# indem dem Elternknoten mitgeteilt wird, dass sein
				# Kind nun `nothing` ist
				set_parent_child(v, nothing, t)
				if !isnothing(v.parent)
					update_height_tree(v.parent) # Hoehe auf dem Pfad zur Wurzel aktualisieren
					if @isdefined(SHOW_BEFORE_REBALANCE) && SHOW_BEFORE_REBALANCE
						println("Before rebalance:")
						println(t)
					end
					rebalance_after_deletion(v.parent, t)
				end
			elseif isnothing(v.left_c) # v hat nur rechtes Kind
				# Neues Kind des Elternknotens ist nun rechtes Kind von v
				set_parent_child(v, v.right_c, t)
				if !isnothing(v.parent)
					update_height_tree(v.parent)
					if @isdefined(SHOW_BEFORE_REBALANCE) && SHOW_BEFORE_REBALANCE
						println("Before rebalance:")
						println(t)
					end
					rebalance_after_deletion(v.parent, t)
				end
			elseif isnothing(v.right_c) # v hat nur linkes Kind
				# Neues Kind des Elternknotens ist nun linkes Kind von v
				set_parent_child(v, v.left_c, t)
				if !isnothing(v.parent)
					update_height(v.parent)
					if @isdefined(SHOW_BEFORE_REBALANCE) && SHOW_BEFORE_REBALANCE
						println("Before rebalance:")
						println(t)
					end
					rebalance_after_deletion(v.parent, t)
				end
			else # v hat sowohl linkes Kind als auch rechtes Kind
				# Bestimme Maximum im linken Teilbaum, uebernehme Wert
				# des Maximums in v und loesche das Maximum aus dem linken
				# Teilbaum (hat garantiert hoechstens ein Kind)
				max_left = max(v.left_c) # Finde Maximum im linken Teilbaum
				v.value = max_left.value # Uebernehme Wert des Maximums
				set_parent_child(max_left, max_left.left_c, t) # Loesche Maximum
				update_height_tree(max_left.parent)
				if @isdefined(SHOW_BEFORE_REBALANCE) && SHOW_BEFORE_REBALANCE
					println("Before rebalance:")
					println(t)
				end
				rebalance_after_deletion(max_left.parent, t)
			end
			return # Schluessel wurde gefunden und geloescht, fertig
		elseif t.key(e) < t.key(v.value) # Suche weiter im linken Teilbaum
			v = v.left_c
		else # t.key(e) > t.key(v.value), Suche weiter im rechten Teilbaum
			v = v.right_c
		end
	end
	nothing
end

function search(e::Any, t::AVLTree) :: Union{Any, Nothing}
	v = t.root
	while !isnothing(v)
		if t.key(e) == t.key(v.value)
			# Schluessel gefunden, das entsprechende Element zurueck
			return v.value
		elseif t.key(e) < t.key(v.value)
			if isnothing(v.left_c)
				return v.value
			end
			v = v.left_c # Suche weiter im linken Teilbaum
		else
			v = v.right_c # Suche weiter im rechten Teilbaum
		end
	end
	return nothing
end

# Hilfsfunktionen
function max(v::AVLTreeNode) :: AVLTreeNode
	# Ermittelt den maximalen Schluessel im Teilbaum v
	current = v
	while !isnothing(current.right_c)
		current = current.right_c
	end
	return current
end

function set_parent_child(v::AVLTreeNode, new_c::Union{AVLTreeNode, Nothing}, t::AVLTree) :: Nothing
	# Setzt das linke Kind des Elternknotens von v, falls v linkes Kind ist
	# bzw. das rechte Kind des Elternknotens von v, falls v rechtes Kind ist
	# auf den neuen Wert `new_c`.
	# Falls es keinen Elternknoten gibt, wird die Wurzel auf `new_c` gesetzt.
	if isnothing(v.parent) # v ist Wurzel
		t.root = new_c
		if !isnothing(new_c)
			t.root.parent = nothing
		end
	elseif v.parent.left_c == v # v ist linkes Kind
		set_left_child(v.parent, new_c)
	else # v ist rechtes Kind
		set_right_child(v.parent, new_c)
	end
	nothing
end

# Ausgabe in Baumdarstellung durch Julia
function Base.show(io::IO, t::AVLTree)
	println(io, "AVL Tree:")
	if isnothing(t.root)
		println("\tEmpty Tree")
	else
		print(io, t.root)
	end
end
function Base.show(io::IO, v::AVLTreeNode)
	show(io, v, 0, true, false)
end
function Base.show(io::IO, v::AVLTreeNode, indent::Int, is_root::Bool, is_left::Bool)
	if !isnothing(v.right_c)
		show(io, v.right_c, indent+1, false, false)
	end
	print(io, repeat("    ", indent), if is_root "-< " elseif is_left "\\- " else "/- " end, v.value)
	factor = balance_factor(v)
	if factor > 1 || factor < -1
		printstyled(io, " (", factor, ")", color=:red, bold=true)
	else
		print(" (", factor, ")")
	end
	println()
	if !isnothing(v.left_c)
		show(io, v.left_c, indent+1, false, true)
	end
end


# Beispielaufrufe
t = make_avl_tree([42, 7, 18, 2])
println(t)
SHOW_BEFORE_REBALANCE = true

for e in [3, 9, 10, 55, 50]
	println("Fuege $e ein:")
	insert(e, t)
	println(t)
end

for e in [-3, 10, 54, 102]
	println("Suche nach $e:", search(e, t))
end
println("")

for e in [18, 9, 2, 3, 7, 10, 50, -5, 42, 55]
	println("Loesche $e:")
	delete(e, t)
	println(t)
end
