#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Trie
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Datentypen nutzen!

	Foliensatz: 07-Mengen-von-Zeichenketten
	Ca. ab Folie(n): 4
"""

# Linked List analog zu Foliensatz 03, Folie 54 laden
include("./../Utils/linked-list.jl")

# Der Bezeichner fuer das leere Zeichen, global definiert
const BOT = '\u22A5'

mutable struct TrieNode
	edgeLabel :: Union{Char, Nothing}
	parent :: Union{TrieNode, Nothing}
	children :: LinkedList
end

function is_leaf(t :: TrieNode) :: Bool
	return empty_list(t.children)
end

mutable struct Trie
	root :: Union{TrieNode, Nothing}
end

function make_trie(elements :: Vector{String}) :: Trie
	t = Trie(nothing)
	for e in elements
		insert(e, t)
	end
	return t
end

function insert(e :: String, t :: Trie) :: Nothing
	is_new = false
	if isnothing(t.root)
		t.root = TrieNode(nothing, nothing, make_list())
		is_new = true
	end
	insert(e, t.root, is_new)
end

function insert(e :: String, t :: TrieNode, is_new = false) :: Nothing
	if isempty(e)
		if (!is_leaf(t) || isnothing(t.parent)) && isnothing(edge(t, BOT))
			# Einfuegen einer neuen Kante mit dem leeren Zeichen
			insert(TrieNode(BOT, t, make_list()), t.children)
		end
	else
		child = edge(t, e[1])
		if !isnothing(child)
			# rekursiver Aufruf mit fehlenden Zeichen
			insert(e[2:end], child)
		else
			if is_leaf(t) && !is_new
				# Einfuegen einer neuen Kante mit dem leeren Zeichen
				insert(TrieNode(BOT, t, make_list()), t.children)
			end

			new_node = TrieNode(e[1], t, make_list())
			# Einfuegen einer neuen Kante mit dem ersten Zeichen
			insert(new_node, t.children)
			# rekursiver Aufruf mit fehlenden Zeichen
			insert(e[2:end], new_node, true)
		end
	end
	nothing
end

function delete(e :: String, t :: Trie) :: Nothing
	if !isnothing(t.root)
		delete(e, t.root)
	end
end

function delete(e :: String, t :: TrieNode) :: Nothing
	if isempty(e)
		if !is_leaf(t)
			bot_child = edge(t, BOT)
			if !isnothing(bot_child)
				t = bot_child
			end
		end
		while !isnothing(t.parent)
			delete(t, t.parent.children)

			# Loesche Knoten mit leerem Zeichen falls keine weiteren Kinder
			if length(t.parent.children) == 1 && !isnothing(edge(t.parent, BOT))
				t.parent.children = make_list()
				break
			end

			# Abbruch, wenn es vor dem Loeschen mehr als ein Kind gab
			if length(t.parent.children) >= 1
				break
			end

			t = t.parent
		end
	else
		child = edge(t, e[1])
		if !isnothing(child)
			delete(e[2:end], child)
		end
	end
	nothing
end

function search(e :: String, t :: Trie) :: Bool
	if isnothing(t.root)
		return false
	else
		return search(e, t.root)
	end
end

function search(e :: String, t :: TrieNode) :: Bool
	if isempty(e)
		if is_leaf(t) || !isnothing(edge(t, BOT))
			return true
		else
			return false
		end
	else
		child = edge(t, e[1])
		if !isnothing(child)
			return search(e[2:end], child)
		else
			return false
		end
	end
end

function edge(t :: TrieNode, c :: Char) :: Union{TrieNode, Nothing}
	for child in t.children
		if child.edgeLabel == c
			return child
		end
	end
	return nothing
end

function Base.show(io :: IO, t :: Trie)
	if isnothing(t.root)
		print(io, "Empty Trie")
		return
	end

	l = [map(c -> c, t.root.children)]
	while !isempty(l)
		c = l[end]
		if isempty(c)
			pop!(l)
		else
			node = pop!(c)
			pre = "    "
			for i = 1:length(l)-1
				pre = pre * (isempty(l[i]) ? "   " : "|  ")
			end
			println(io, string(pre, "|- ", node.edgeLabel))
			if !empty_list(node.children)
				push!(l, map(c -> c, node.children))
			end
		end
	end
end

# Beispielaufrufe
println("Erstelle Trie mit ape, apple, organ, organism")
t = make_trie(["ape", "apple", "organ", "organism"])
println(t)

println("Suche nach ape: " * string(search("ape", t)))
println("Suche nach organ: " * string(search("organ", t)))
println("Suche nach apricot: " * string(search("apricot", t)))
println("Suche nach org: " * string(search("org", t)))

println("Fuege apricot ein")
insert("apricot", t)
println(t)
println("Suche nach apricot: " * string(search("apricot", t)))

println("Fuege org ein")
insert("org", t)
println(t)
println("Suche nach org: " * string(search("org", t)))

println("Loesche ape")
delete("ape", t)
println("Suche nach ape: " * string(search("ape", t)))
println(t)

println("Loesche organ")
delete("organ", t)
println("Suche nach organ: " * string(search("organ", t)))
println(t)