#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Patricia Trie
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Datentypen nutzen!

	Foliensatz: 07-Mengen-von-Zeichenketten
	Ca. ab Folie(n): 23
"""

# Linked List analog zu Foliensatz 03, Folie 54 laden
include("./../Utils/linked-list.jl")

# Der Bezeichner fuer das leere Zeichen, global definiert
const BOT = "\u22A5"

mutable struct PatriciaTrieNode
	edgeLabel :: Union{String, Nothing}
	parent :: Union{PatriciaTrieNode, Nothing}
	children :: LinkedList
end

function is_leaf(t :: PatriciaTrieNode) :: Bool
	return empty_list(t.children)
end

function is_root(t :: PatriciaTrieNode) :: Bool
	return isnothing(t.parent)
end

function Base.:(==)(t1 :: PatriciaTrieNode, t2 :: PatriciaTrieNode) :: Bool
	if isnothing(t1.parent)
		return isnothing(t2.parent) && t1.edgeLabel == t2.edgeLabel
	else
		return t1.parent.edgeLabel == t2.parent.edgeLabel &&
			t1.edgeLabel == t2.edgeLabel
	end
end

mutable struct PatriciaTrie
	root :: Union{PatriciaTrieNode, Nothing}
end

function make_patricia_trie(elements :: Vector{String}) :: PatriciaTrie
	t = PatriciaTrie(nothing)
	for e in elements
		insert(e, t)
	end
	return t
end

function insert(e :: String, t :: PatriciaTrie) :: Nothing
	if isnothing(t.root)
		t.root = PatriciaTrieNode(nothing, nothing, make_list())
	end
	insert(e, t.root)
end

function insert(e :: String, t :: PatriciaTrieNode) :: Nothing
	if isempty(e) && is_root(t)
		insert(PatriciaTrieNode(BOT, t, make_list()), t.children)
		return
	end

	pre_edge = prefix_edge(t, e)
	if !isnothing(pre_edge)
		(c, prefix) = pre_edge
		suffix = e[length(prefix)+1:end]
		if isempty(suffix)
			suffix = BOT
		end

		if prefix == c.edgeLabel && suffix != BOT
			# rekursiver Aufruf mit fehlenden Zeichen
			insert(suffix, c)
		else
			new_label = c.edgeLabel[length(prefix)+1:end]
			if empty_list(c.children) # Unter das Kind anfuegen
				insert(PatriciaTrieNode(new_label, c, make_list()), c.children)
				insert(PatriciaTrieNode(suffix, c, make_list()), c.children)
			else # Aufteilen des Knotens
				old_childs = c.children
				c.children = make_list()
				insert(PatriciaTrieNode(suffix, c, make_list()), c.children)
				new_node = PatriciaTrieNode(new_label, c, old_childs)
				insert(new_node, c.children)
				for cc in old_childs
					cc.parent = new_node
				end
			end
			c.edgeLabel = prefix
		end
	else
		if is_leaf(t) && !is_root(t)
			insert(PatriciaTrieNode(BOT, t, make_list()), t.children)
		end
		insert(PatriciaTrieNode(e, t, make_list()), t.children)
	end
	nothing
end

function delete(e :: String, t :: PatriciaTrie) :: Nothing
	if !isnothing(t.root)
		delete(e, t.root)
	end
end

function delete(e :: String, t :: PatriciaTrieNode) :: Nothing
	if isempty(e)
		if !is_leaf(t)
			pre_edge = prefix_edge(t, BOT)
			if !isnothing(pre_edge)
				(t, _) = pre_edge
			end
		end
		while !isnothing(t.parent)
			delete(t, t.parent.children)

			num_childs = length(t.parent.children)
			# Loesche Knoten mit leerem Zeichen falls keine weiteren Kinder
			if num_childs == 1 && !isnothing(prefix_edge(t.parent, BOT))
				t.parent.children = make_list()
				break
			end

			# Abbruch, wenn es vor dem Loeschen mehr als ein Kind gab
			if num_childs >= 1
				break
			end

			t = t.parent
		end
		if length(t.parent.children) == 1 && !is_root(t.parent)
			c = first(t.parent.children)
			c.parent = t.parent.parent
			delete(t.parent, t.parent.parent.children)
			insert(c, t.parent.parent.children)
			c.edgeLabel = t.parent.edgeLabel * c.edgeLabel
		end
	else
		pre_edge = prefix_edge(t, e)
		if !isnothing(pre_edge)
			(c, prefix) = pre_edge
			if length(prefix) != length(e) || e == c.edgeLabel
				delete(e[length(prefix)+1:end], c)
			end
		end
	end
	nothing
end

function search(e :: String, t :: PatriciaTrie) :: Bool
	if isnothing(t.root)
		return false
	else
		return search(e, t.root)
	end
end

function search(e :: String, t :: PatriciaTrieNode) :: Bool
	if isempty(e)
		if is_leaf(t) || !isnothing(prefix_edge(t, BOT))
			return true
		else
			return false
		end
	else
		pre_edge = prefix_edge(t, e)
		if !isnothing(pre_edge)
			(c, prefix) = pre_edge
			if length(prefix) == length(e) && e != c.edgeLabel
				# Kantenbeschriftung enthaelt mehr Zeichen als gesucht
				return false
			else
				return search(e[length(prefix)+1:end], c)
			end
		else
			return false
		end
	end
end

function prefix_edge(t :: PatriciaTrieNode, s :: String) :: 
		Union{Tuple{PatriciaTrieNode, String}, Nothing}
	for child in t.children
		same_prefix = longest_prefix_match(s, child.edgeLabel)
		if !isempty(same_prefix)
			return (child, same_prefix)
		end
	end
	return nothing
end

function longest_prefix_match(s1 :: String, s2 :: String) :: String
	count = 0
	for i = 1:minimum([lastindex(s1), lastindex(s2)])
		if s1[i] == s2[i]
			count += 1
		else
			break
		end
	end
	return s1[1:count]
end

function Base.show(io :: IO, t :: PatriciaTrie)
	if isnothing(t.root)
		print(io, "Empty PatriciaTrie")
		return
	end

	l = [map(c -> c, t.root.children)]
	while !isempty(l)
		c = l[end]
		if isempty(c)
			pop!(l)
		else
			node = pop!(c)
			pre = "    "
			for i = 1:length(l)-1
				pre = pre * (isempty(l[i]) ? "   " : "|  ")
			end
			println(io, string(pre, "|- ", node.edgeLabel))
			if !empty_list(node.children)
				push!(l, map(c -> c, node.children))
			end
		end
	end
end


# Beispielaufrufe
println("Erstelle PatriciaTrie mit ape, apple, organ, organism")
t = make_patricia_trie(["ape", "apple", "organ", "organism"])
println(t)

println("Suche nach ape: " * string(search("ape", t)))
println("Suche nach organ: " * string(search("organ", t)))
println("Suche nach organization: " * string(search("organization", t)))
println("Suche nach org: " * string(search("org", t)))

println("Fuege organization ein")
insert("organization", t)
println(t)
println("Suche nach organization: " * string(search("organization", t)))

println("Fuege org ein")
insert("org", t)
println(t)
println("Suche nach org: " * string(search("org", t)))

println("Loesche apple")
delete("apple", t)
println(t)
println("Suche nach apple: " * string(search("apple", t)))

println("Loesche organ")
delete("organ", t)
println(t)
println("Suche nach organ: " * string(search("organ", t)))