#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Union Find mit Pfadkompression
		White-Box API

	Foliensatz: 06-Disjukte Mengen
	Operationen nach ca. Folie(n): 17
	Umsetzung nach ca. Folie(n): 21, 27
"""

# Linked List analog zu Foliensatz 03, Folie 54 laden
include("./../Utils/linked-list.jl")

mutable struct UnionFindElement 
	value :: Any
	parent :: Union{Nothing, UnionFindElement}
	rank :: Int
		# Nach find() mit Pfadkompression ist der Rang nur noch
		# eine obere Schranke, da Updates am Rang nur bei union() stattfinden,
		# siehe: Robert E. Tarjan, Jan van Leeuwen. Worst-case analysis of
		# set union algorithms, Journal of the ACM 31 (2), S. 245–281, 1984
end 

# Zugriff auf Attribute
function value(x::UnionFindElement) :: Any
	return x.value
end

function rank(x::UnionFindElement) :: Int
	return x.rank
end

# Operationen 
function make_set(e::Any) :: UnionFindElement
	return UnionFindElement(e, nothing, 0)
end

function find(x::UnionFindElement) :: UnionFindElement
	visited = make_list()
	while !isnothing(x.parent)
		insert(x, visited)
		x = x.parent
	end
	representant = x

	for x_ in visited # Pfadkompression 
		x_.parent = representant 
	end
	
	return representant
end

function union(x::UnionFindElement, x_::UnionFindElement) :: Nothing
	x = find(x)
	x_ = find(x_)

	if x != x_
		if rank(x) >= rank(x_)
			x_.parent = x 
		else 
			x.parent = x_
		end

		if rank(x) == rank(x_)
			x.rank += 1
		end
	end

	nothing
end

# Ausgabe 
function Base.show(io::IO, x::UnionFindElement) 
	str = ""
	while !isnothing(x)
		str *= string(value(x), " (", x.rank ,") -> ")
		x = x.parent
	end
	print(io, str[begin:end-4])
end
function Base.show(io::IO, elems::Vector{UnionFindElement}) 
	println(io, "Union Find Elemente [Element (Rang) -> ... -> Repraesentant]:")
	for x in elems
		println(io, "\t", x)
	end
end

# Beispielaufrufe
elements = [1,2,3,4,5,6,7,8]
element_sets = [make_set(e) for e in elements]
println(element_sets)

println("union(2, 3), union(4, 5), union(5, 6)\n")
union(element_sets[2], element_sets[3])
union(element_sets[4], element_sets[5])
union(element_sets[5], element_sets[6])
println(element_sets)

println("union(1,2), union(1,4)\n")
union(element_sets[1], element_sets[2])
union(element_sets[1], element_sets[4])
println(element_sets)

union(element_sets[7], element_sets[8])
println("union(7, 8), find(6) = ", find(element_sets[6]), "\n")
println(element_sets)

println("union(5, 2)\n")
union(element_sets[5], element_sets[2])
println(element_sets)

println("union(7, 2)\n")
union(element_sets[7], element_sets[2])
println(element_sets)