#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Hashing mit Kollisionslisten (geschlossene Adressierung)
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Datentypen nutzen!

	Foliensatz: 09-Assoziation-von-Objekten
	Ca. ab Folie(n): 17
"""

# Linked List analog zu Foliensatz 03, Folie 54 laden
include("./../Utils/linked-list.jl")

mutable struct HashtableCollisionList
	internal_repr::Array{LinkedList} # Interne Repraesentation
	h::Function # Interne Hashfunktion
	n::Int # Anzahl an gespeicherten Elementen fuer dynamische Reallokation
	key::Function # Benutzerdefinierte Funktion Element -> Schluessel
end

function HashtableCollisionList(key=identity, length=10)::HashtableCollisionList
	T = [make_list() for _ = 1:length]
	h = function(u)
		p = generate_prime(length)
		return (key(u) % p) % length + 1
	end
	return HashtableCollisionList(T, h, 0, key)
end

function insert(k, e, d::HashtableCollisionList)
	T = d.internal_repr
	insert((k, e), T[d.h(k)])
	d.n += 1
	if d.n > convert(Int, floor(length(T) / 2))
		resize(d, true) # Verdopple Tabellengroesse
	end
end

function delete(k, d::HashtableCollisionList)
	e = lookup(k, d)
	if !isnothing(e)
		T = d.internal_repr
		delete((k, e), T[d.h(k)])
		d.n -= 1
		if d.n <= convert(Int, floor(length(T) / 4))
			resize(d, false) # Halbiere Tabellengroesse
		end
	end
end

function lookup(k, d::HashtableCollisionList)::Any
	T = d.internal_repr
	for (k_, e) in T[d.h(k)]
		if k == k_
			return e
		end
	end
	return nothing
end

function resize(d::HashtableCollisionList, enlarge::Bool)
	T = d.internal_repr
	if enlarge
		new_length = 2*length(T)
	else
		new_length = convert(Int, floor(length(T) / 2))
	end
	T_new = [make_list() for _ = 1:new_length]

	# Hashfunktion an die neue Groesse angepassen
	d.h = function(u)
		p = generate_prime(new_length)
		return (d.key(u) % p) % new_length + 1
	end

	# Elemente in neue Tabelle uebertragen
	for i = 1:length(T)
		for (k, e) in T[i]
			insert((k, e), T_new[d.h(k)])
		end
	end

	d.internal_repr = T_new
end

function generate_prime(k::Int)::Int
	for m = k^3:(k+1)^3
		if is_prime(m, k)
			return m
		end
	end
	return -1
end

function is_prime(m::Int, k::Int)::Bool
	for i = 2:convert(Int, floor(sqrt((k+1)^3)))
		if m % i == 0
			return false
		end
	end
	return true
end

function Base.show(io::IO, d::HashtableCollisionList)
	println("Dynamische Hashtabelle mit Kollisionslisten und $(d.n) Elementen:")
	T = d.internal_repr
	for index = 1:length(T)
		print(io, string(lpad(index, length(string(length(T))), "0"), ": "))
		elems_str = ""
		for (k, e) in T[index]
			elems_str *= string((k, e), " -> ")
		end
		println(io, elems_str[1:end-4])
	end
end

# Beispielaufrufe
ht = HashtableCollisionList(identity, 2)
println(ht)
insert(11, "A", ht)
println(ht)
insert(15, "B", ht)
println(ht)
insert(22, "C", ht)
println(ht)

println(string("Suche nach 11: ", lookup(11, ht)))
println(string("Suche nach  1: ", lookup(1, ht)))
println(string("Suche nach 22: ", lookup(22, ht), "\n"))

delete(11, ht)
println(ht)
delete(22, ht)
println(ht)

println("===\n")
# Beispielaufrufe mit benutzerdefiniertem Datenyp und eigener key-Funktion
mutable struct Student
	matr_nr::Int
	name::String
end

students = [
	Student(123, "Alice"),
	Student(456, "Bob"),
	Student(789, "Charlie"),
]

key = function(e::String)
	val = 0
	for (pos, char) in enumerate(e)
		val += pos * Int(char)
	end
	return val
end

ht_ = HashtableCollisionList(key, 5)
for s in students
	insert(s.name, s, ht_)
	println(ht_)
end

println(string("Suche nach Bob: ", lookup("Bob", ht_)))
println(string("Suche nach Charly: ", lookup("Charly", ht_)))
println(string("Suche nach Charlie: ", lookup("Charlie", ht_), "\n"))

for s in students[1:2]
	delete(s.name, ht_)
	println(ht_)
end