#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Henrik Strunck
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Hashing mit Sondieren (Offene Adressierung)
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Datentypen nutzen!

	Foliensatz: 09-Assoziation-von-Objekten
	Ca. ab Folie(n): 25
"""

# Bezeichner für einen gelöschten Eintrag in der Hashtabelle
const EMPTY = "\u03B5"

mutable struct HashtableOpenAddressing
	internal_repr :: Array{Any}
	h :: Function # Hashfunktion
	f :: Function # Sondierungsfunktion
	n :: Int
	key :: Function
end

function HashtableOpenAddressing(f, key=identity, length=10) :: HashtableOpenAddressing
	T = Array{Any}(nothing, length)
	h = function(u)
		p = generate_prime(length)
		pos = (key(u) % p) % length
		return pos < 0 ? pos + length : pos + 1
	end
	return HashtableOpenAddressing(T, h, f, 0, key)
end

function insert(k, e, d::HashtableOpenAddressing) :: Nothing
	T = d.internal_repr
	i = d.h(k)
	step = 1
	while !isnothing(T[i]) && T[i] != EMPTY && T[i][1] != k
		i = (d.h(k) + d.f(step)) % length(T)
		step += 1
	end
	T[i] = (k,e)

	d.n += 1
	if d.n > convert(Int, floor(length(T) / 2))
		resize(d, true) # Verdopple Tabellengroesse
	end

	nothing
end

function delete(k, d::HashtableOpenAddressing) :: Nothing
	T = d.internal_repr
	i = d.h(k)
	step = 1
	while !isnothing(T[i]) && (T[i] == EMPTY || T[i][1] != k)
		i = (d.h(k) + d.f(step)) % length(T)
		if step == length(T)
			break
		end
		step += 1
	end
	if !isnothing(T[i]) && T[i][1] == k
		T[i] = EMPTY

		d.n -= 1
		if d.n <= convert(Int, floor(length(T) / 4))
			resize(d, false) # Halbiere Tabellengroesse
		end
	end

	nothing
end

function lookup(k, d::HashtableOpenAddressing) :: Union{Any, Nothing}
	T = d.internal_repr
	i = d.h(k)
	step = 1
	while !isnothing(T[i]) && (T[i] == EMPTY || T[i][1] != k)
		i = (d.h(k) + d.f(step)) % length(T)
		if step == length(T)
			break
		end
		step += 1
	end
	if isnothing(T[i]) || T[i] == EMPTY || T[i][1] != k
		return nothing
	else
		return T[i][2]
	end
end

function resize(d::HashtableOpenAddressing, enlarge::Bool) :: Nothing
	T = d.internal_repr
	if enlarge
		new_length = 2 * length(T)
	else
		new_length = convert(Int, floor(length(T) / 2))
	end
	T_new = Array{Any}(nothing, new_length)

	# Hashfunktion an die neue Groesse anpassen
	d.h = function(u)
		p = generate_prime(new_length)
		pos = (d.key(u) % p) % new_length
		return pos < 0 ? pos + new_length : pos + 1
	end

	# Elemente in neue Tabelle uebertragen
	for i = 1:length(T)
		if !isnothing(T[i]) && T[i] != EMPTY
			T_new[d.h(T[i][1])] = T[i]
		end
	end

	d.internal_repr = T_new

	nothing
end

function generate_prime(k::Int) :: Int
	for m = k^3:(k + 1)^3
		if is_prime(m, k)
			return m
		end
	end
	return -1
end

function is_prime(m::Int, k::Int) :: Bool
	for i = 2:convert(Int, floor(sqrt((k + 1)^3)))
		if m % i == 0
			return false
		end
	end
	return true
end

function Base.show(io::IO, d::HashtableOpenAddressing)
	println("Dynamische Hashtabelle mit Sondierung und $(d.n) Elementen:")
	T = d.internal_repr
	for i = 1:length(T)
		print(io, string(lpad(i, length(string(length(T))), "0"), ": "))
		println(io, T[i])
	end
end

# Sondierungsfunktionen
# Der Parameter i gibt den Sondierungsschritt an (0, 1, ...)
function linear_probing(i::Int) :: Int
	return i
end

function quadratic_probing(i::Int) :: Int
	return i * i
end


# Beispielaufrufe
println("--- Lineares Sondieren ---")
ht = HashtableOpenAddressing(linear_probing)
println(ht)
println("insert(42, \"Apple\", ht)")
insert(42, "Apple", ht)
println(ht)
println("insert(18, \"Pear\", ht)")
insert(18, "Pear", ht)
println(ht)
println("insert(4, \"Orange\", ht)")
insert(4, "Orange", ht)
println(ht)
println("insert(14, \"Ginger\", ht)")
insert(14, "Ginger", ht)
println(ht)
println("insert(24, \"Pineapple\", ht)")
insert(24, "Pineapple", ht)
println(ht)

println(string("Suche nach 24: ", lookup(24, ht)))
println(string("Suche nach 18: ", lookup(18, ht)))
println(string("Suche nach -5: ", lookup(-5, ht), "\n"))

println("delete(24, ht)")
delete(24, ht)
println(ht)
println("delete(4, ht)")
delete(4, ht)
println(ht)
println("delete(18, ht)")
delete(18, ht)
println(ht)
println("delete(77, ht)")
delete(77, ht)
println(ht)


println("--- Quadratisches Sondieren ---")
ht = HashtableOpenAddressing(quadratic_probing)
println(ht)
println("insert(42, \"Apple\", ht)")
insert(42, "Apple", ht)
println(ht)
println("insert(18, \"Pear\", ht)")
insert(18, "Pear", ht)
println(ht)
println("insert(4, \"Orange\", ht)")
insert(4, "Orange", ht)
println(ht)
println("insert(14, \"Ginger\", ht)")
insert(14, "Ginger", ht)
println(ht)
println("insert(24, \"Pineapple\", ht)")
insert(24, "Pineapple", ht)
println(ht)
