#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	A* Algorithmus
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Algorithmen nutzen!

	Foliensatz: 10-Graphen
	Ca. ab Folie(n): 100
"""

# Datenstruktur laden 
USE_GRAPH_PLOTS = true 
include("./../Utils/graph.jl")
# Prioritaetswarteschlange laden
include("./../Utils/binary-heap-pq.jl")
# Stack laden
include("./../Utils/stack.jl")

mutable struct AStarNode
	g::Union{Nothing, Float32}
	f::Union{Nothing, Float32}
	parent::Union{Nothing, AStarNode}
	value::Any

	AStarNode(value) = new(nothing, nothing, nothing, value)
end
function string(v::AStarNode)
	return v.value
end
Base.hash(v::AStarNode) = hash(string(v)) # used for mapping of elements to nodes in PQ and Set()
Base.isequal(u::AStarNode, v::AStarNode) = isequal(string(u), string(v))
Base.:(==)(u::AStarNode, v::AStarNode) = isequal(u, v)

# Eingabe: Startknoten s, Zielknoten goal, Schaetzer(-funktion) h und Graph g
function a_star(s::AStarNode, goal::AStarNode, h::Function, g::Graph)
	s.g = 0 			# Kein Weg zum Startknoten
	s.f = s.g + h(s)	# Schaetzung durchfuehren
	s.parent = s 		# Eltern zeigen auf sich selbst

	# PQ nutzt Schaetzungen; Menge der expandierten Knoten
	pq = build_pq([s], (u) -> u.f, (u, new_key) -> u.f = new_key)
	expanded = Set{AStarNode}() 	# Nutzt hash() intern

	while !empty_pq(pq)
		u = delete_min(pq)
		push!(expanded, u)
		if u == goal 
			return path(u) 	# Loesungspfad (rueckwaerts) erstellen
		end

		for (_, v, cost) in edges_out(u, g)
			if v ∉ expanded 	# Knoten noch nicht besucht?
				if v ∉ pq 	# Neuen Knoten gesehen?
					v.g = u.g + cost 	# Wegkosten
					v.f = v.g + h(v)	# Schaetzer bestimmen 
					v.parent = u		# Eltern
					insert(v, pq)
				else 		# Bekannter Knoten, neuer Weg dorthin
					new_g = u.g + cost
					if new_g < v.g # Neuer Weg kuezer?
						decrease_key(v, pq, v.g - new_g ) # Neue Position in PQ (aktualisiert auch v.f)
						v.parent = u # Eltern und Wegkosten anpassen
						v.g = new_g
					end
				end
			elseif u.g + cost < v.g # Kuerzerer Weg zu bereits expandiertem Knoten gefunden
				propagate(u, v, u.g + cost, expanded, pq, g)
			end
		end
	end
end

function path(u::AStarNode)
	# Konstruiere Pfad p ausgehend vom Zielknoten u 
	p = make_stack()
	while u.parent != u # Startknoten hat sich selbst als Eltern
		push(u, p)
		u = u.parent
	end
	push(u, p) # Startknoten anfuegen
	return p
end

function propagate(u::AStarNode, v::AStarNode, new_g::Float32, expanded::Set{AStarNode}, pq::PQ, g::Graph)
	# Propagiere neue Kosten new_g fuer Knoten v in die Nachfolger von v
	if v ∉ expanded
		if new_g < v.g # Kuerzerer Weg zu noch nicht expandiertem Knoten
			decrease_key(v, pq, v.g - new_g ) # Neue Position in PQ 
			v.parent = u 
			v.g = new_g
		end
	else # Knoten schon expandiert
		if new_g < v.g 	# Kuerzerer Weg gefunden
			v.parent = u 	# Neuen Weg uebernehmen
			v.g = new_g
			# Weiter propagieren 
			for (_, w, cost) in edges_out(v, g)
				propagate(v, w, new_g + cost, expanded, pq, g)
			end
		end
	end
end


function a_star_print(g::Graph, p::Stack, h::Function)
	println(g)
	println("\tKuerzester Pfad: ", p)
	println("\tA* Werte:")
	nodes_sorted = sort(collect(nodes(g)), by=(x) -> (x.value))
	for u in nodes_sorted
		print("\t\t$(u.value): ")
		print("h($(u.value)) = $(lpad(h(u), 4, " ")), ")
		print("g[$(u.value)] = $(lpad(u.g, 4, " ")), ")
		print("f[$(u.value)] = $(lpad(u.f, 4, " ")), ")
		println("parent[$(u.value)] = $(isnothing(u.parent) ? "nothing" : u.parent.value)")
	end
end

# Beispielaufrufe
# 	Graph anlog zu Folien 110-121
s, a, b, c = AStarNode("s"), AStarNode("a"), AStarNode("b"), AStarNode("c")
d, e, f = AStarNode("d"), AStarNode("e"), AStarNode("f")
graph_i = DirectedWeightedGraph([
		s => d, d => e, e => f, s => a, a => b, b => c, c => f
	], [ 2, 3, 2, 1.5, 2, 3, 4 ], # Kantengewichte
	"g-AStarGraph-i"
)
estimations_i = Dict(s=>6, a=>4, b=>2, c=>4, d=>4.5, e=>2, f=>0)

path_i = a_star(s, f, (u) -> estimations_i[u], graph_i)
a_star_print(graph_i, path_i, (u) -> estimations_i[u])

#	Weiterer Graph
A, B, C, D = AStarNode("A"), AStarNode("B"), AStarNode("C"), AStarNode("D")
E, F, G, H = AStarNode("E"), AStarNode("F"), AStarNode("G"), AStarNode("H")
graph_ii = WeightedGraph([
		A => G, A => B, B => H, B => E, B => D, B => C,
		C => D, D => E, E => F, F => H, F => G, G => H
	], [ 
		 2,  6,  7,  8, 10,  4,
		 4, 11,  4, 15,  3, 10
	], 
	"g-AStarGraph-ii"
)
estimations_ii = Dict(A=>12, B=>6, C=>4, D=>0, E=>8, F=>15, G=>10, H=>15)

path_ii = a_star(A, D, (u) -> estimations_ii[u], graph_ii)
a_star_print(graph_ii, path_ii, (u) -> estimations_ii[u])