#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Ford-Fulkerson
		mit Edmonds-Karp zur Bestimmung des flusserhöhenden Pfades

	Foliensatz: 10-Graphen
	Ca. ab Folie(n): 212, 226
"""

# Datenstruktur laden 
USE_GRAPH_PLOTS = true 
include("./../Utils/graph.jl")
# Queue analog zu Foliensatz 03, Folie 70 laden
include("./../Utils/queue.jl")

# make dict of edges return default value of 0 for unknown edges (edge tuple key)
Base.getindex(d::Dict, uv::Tuple{Any, Any}) = get(d, uv, 0)

function ford_fulkerson(s::Any, t::Any, g::Graph)
	rest_capacities = Dict{Tuple{Any, Any}, Number}() # default_value=0 (for undefined key)
	for (u, v, c) in edges(g)
		rest_capacities[u, v] = c
	end
	max_flow = 0

	argumenting_path = flow_argumenting_path(s, t, g, rest_capacities)
	while !isnothing(argumenting_path)
		path_edges, path_capacity = argumenting_path
		for (u, v) in path_edges
			rest_capacities[u, v] -= path_capacity
			rest_capacities[v, u] += path_capacity
		end
		max_flow += path_capacity

		argumenting_path = flow_argumenting_path(s, t, g, rest_capacities)
	end

	return max_flow, rest_capacities
end

# Edmonds-Karp Algorithmus zur Wahl von "guenstigen" Flusserhoehenden Pfaden
function flow_argumenting_path(s::Any, t::Any, g::Graph, rest_capacities::Dict{Tuple{Any, Any}, Number})
	parents = Dict{Any,Any}()
	
	# do a bfs
	q = make_queue([s])
	parents[s] = s
	while !empty_queue(q)
		u = dequeue(q) 
		# Wir brauchen die ausgehenden Kanten und dazu auch
		#	noch die eingehenden Kanten, denn letztere koennten
		#	eine entgegengesetzte Restkapazitaet beinhalten.
		edges = [v for (_,v) in edges_out(u, g)]
		append!(edges, (v for (v, _) in edges_in(u, g)))
		for v in edges
			if !haskey(parents, v) && rest_capacities[u, v] > 0
				enqueue(v, q)
				parents[v] = u
			end
		end
	end

	if !haskey(parents, t)
		return nothing 		# no path reaching t found
	else
		# get the path and its capacity
		path_capacity = typemax(Int)
		path = []
		current = t
		while current != s
			path_capacity = minimum((path_capacity, rest_capacities[parents[current], current]))
			insert!(path, 1, (parents[current], current))
			current = parents[current]
		end
		return path, path_capacity
	end
end

function ff_print(g::Graph, rest_capacities::Dict{Tuple{Any, Any}, Number})
	println(g)
	println("\tFord Fulkerson edges:")
	println("\t\t edge : flow / capacity")
	for (u, v, capacity) in sort(collect(edges(g)), by=(e) -> string(e[1], e[2]))
		println("\t\t$u - $v : ",
			lpad(capacity - rest_capacities[u, v], 4, " "), " / ",
			lpad(capacity, 4, " ")
		)
	end
end

# Beispielaufrufe
#	Graph analog zu Folie 206ff
graph_i = DirectedWeightedGraph([
		"s" => "a", "s" => "c", "a" => "c", "a" => "b", "b" => "c", "b" => "t", 
		"c" => "a", "c" => "d", "d" => "b", "d" => "t"
	], [
		    16,     13,     10,     12,      9,     20,
		     4,     14,      7,      4   
	], "k-FordFulkersonGraph-i"
)
max_flow_i, rest_capacities_i = ford_fulkerson("s", "t", graph_i)
ff_print(graph_i, rest_capacities_i)
println("\tMax flow: $max_flow_i")

# 	Graphen aus Uebung
graph_ii = DirectedWeightedGraph([
		"s" => "a", "a" => "b", "a" => "d", "b" => "t",
		"s" => "c", "c" => "b", "c" => "d", "d" => "t"
	], [
		8, 3, 3, 6, 
		6, 3, 6, 8
	], "k-FordFulkersonGraph-ii"
)
max_flow_ii, rest_capacities_ii = ford_fulkerson("s", "t", graph_ii)
ff_print(graph_ii, rest_capacities_ii)
println("\tMax flow: $max_flow_ii")

graph_iii = DirectedWeightedGraph([
		"s" => "a", "s" => "b", "a" => "d", "a" => "c", "b" => "f", "c" => "b",
		"c" => "h", "d" => "g", "d" => "e", "e" => "g", "e" => "c", "f" => "h",
		"g" => "t", "h" => "g", "h" => "t"
	], [
		10, 8, 7, 6, 6, 9,
		 4, 4, 2, 4, 8, 5,
		 8, 6, 8
	], "k-FordFulkersonGraph-iii"
)
max_flow_iii, rest_capacities_iii = ford_fulkerson("s", "t", graph_iii)
ff_print(graph_iii, rest_capacities_iii)
println("\tMax flow: $max_flow_iii")