#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

# Datenstruktur fuer MinMax
struct MinMaxLeaf 
	# Ein Blatt des Baumes repraesentiert ein Ergebnis

	#	Wert dieses Ergebnisses
	value :: Int
end
mutable struct MinMaxNode
	# Ein innerer Knoten des Baumen repraesentiert einen Zustand

	#	Falls `max`, dann Aktion des Spielers, anderfalls des Gegenspielers
	is_max :: Bool
	#	Moegliche Nachfolgezustaende, die durch durch Aktionen von diesem Zustand aus zu erreichen sind
	subsequent_states :: Union{Vector{MinMaxNode}, Vector{MinMaxLeaf}}
	#	Wert des Ergebnis von dieser Aktion aus (zu berechnen durch den Algo.)
	v :: Union{Int, Nothing}

	MinMaxNode(im, a) = new(im, a, nothing)	
end
struct MinMaxTree
	root :: MinMaxNode
end

# Erzeugen von MinMax Baumen
function build_subtree(i::Int, j::Int, is_max::Bool, branching_factor::Int, outcomes::Vector{Int}) :: MinMaxNode
	if j-i > branching_factor
		ranges = Int((j-i) / branching_factor)
		sub_nodes = [
			build_subtree(i+(b*ranges), i+((b+1)*ranges), !is_max, branching_factor, outcomes)
				for b in 0:branching_factor-1
		]
		return MinMaxNode(is_max, sub_nodes)
	else # Blaetter des Baumes erstellen
		return MinMaxNode(is_max, [
			MinMaxLeaf(o) for o in outcomes[i:j-1]
		])
	end
end
function build_tree(depth::Int, branching_factor::Int, outcomes::Vector{Int}) :: MinMaxTree
	needed_leafs = branching_factor ^ depth
	if length(outcomes) != needed_leafs
		throw(ArgumentError("Needs $needed_leafs outcomes for depth=$depth and branching_factor=$branching_factor !"))
	end

	root = build_subtree(1, needed_leafs+1, true, branching_factor, outcomes)	
	return MinMaxTree(root)
end

# Ausgabe von MinMax Baeumen
function Base.show(io::IO, t::MinMaxTree)
	println(io, "Min Max Tree:")
	if isnothing(t.root)
		println("\tEmpty Tree")
	else
		print(io, t.root)
	end
end
function Base.show(io::IO, s::MinMaxNode, indent::Int=0)
	print(io, repeat("   ", indent))
	printstyled(io, if s.is_max "<" else ">" end, bold=true)
	println( " -|", if isnothing(s.v) "" else string(" ($(s.v))") end)
	for su in s.subsequent_states
		show(io, su, indent+1)
	end
end
function Base.show(io::IO, s::MinMaxLeaf, indent::Int=0)
	print(io, repeat("   ", indent), "- ")
	printstyled(io, s.value, bold=true)
	println()
end

"""
	MinMax (Rekursive Implementation)

	Foliensatz: 11-Suchraumbeschneidung
	Ca. ab Folie(n): 19
"""

value(state::MinMaxLeaf) = state.value 
is_terminal(state::MinMaxLeaf) = true
is_terminal(state::MinMaxNode) = false

function minmax_decision(t::MinMaxTree)
	# Bestimme den besten naechsten Schritt 

	#	Berechne alle Werte
	max_value(t.root)
	#	bestimme Aktion/ Zustand zu bestem Wert
	v, i = findmax((s) -> s.v, t.root.subsequent_states)
	return (i, v, t.root.subsequent_states[i])
end

function min_value(state::Union{MinMaxNode, MinMaxLeaf})
	if is_terminal(state)
		return value(state)
	end

	v = typemax(Int)
	for sub_state in state.subsequent_states
		v = minimum((v, max_value(sub_state)))
	end
	state.v = v
	return v
end

function max_value(state::Union{MinMaxNode, MinMaxLeaf})
	if is_terminal(state)
		return value(state)
	end

	v = typemin(Int)
	for sub_state in state.subsequent_states
		v = maximum((v, min_value(sub_state)))
	end
	state.v = v
	return v
end

# Beispielaufrufe
#	Tiefe = 2, Aktionen = 3, Ergebnisse = [...]
#	Analog zu Folie 14f.
t = build_tree(2, 3, [3, 12, 8, 2, 4, 6, 14, 5, 2])
(a_index, a_value, a_node) = minmax_decision(t)
println("Beste Aktion:\n\tAktion Nr. $a_index mit dem Wert $a_value:")
println(t)

# 	Groesserer Baum 
t = build_tree(3, 3, [12, 3, -13, -4, -20, -13, 8, 2, -17, -8, -8, -2, -8, -13, 6, 15, 17, 15, -10, -2, -1, 10, 7, -5, -4, 10, 14])
(a_index, a_value, a_node) = minmax_decision(t)
println("Beste Aktion:\n\tAktion Nr. $a_index mit dem Wert $a_value:")
println(t)