#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

# Datenstruktur fuer MinMax
struct MinMaxLeaf 
	# Ein Blatt des Baumes repraesentiert ein Ergebnis

	#	Wert dieses Ergebnisses
	value :: Int
end
mutable struct MinMaxNode
	# Ein innerer Knoten des Baumen repraesentiert einen Zustand

	#	Falls `max`, dann Aktion des Spielers, anderfalls des Gegenspielers
	is_max :: Bool
	#	Moegliche Nachfolgezustaende, die durch durch Aktionen von diesem Zustand aus zu erreichen sind
	subsequent_states :: Union{Vector{MinMaxNode}, Vector{MinMaxLeaf}}

	#	Wert des Ergebnis von dieser Aktion aus (zu berechnen durch den Algo.)
	v :: Union{Int, Nothing}
	#	Alpha, Beta Werte an diesem Zustand
	alpha :: Union{Int, Nothing} 
	beta :: Union{Int, Nothing}

	MinMaxNode(im, su) = new(im, su, nothing, nothing, nothing)	
end
struct MinMaxTree
	root :: MinMaxNode
end

# Erzeugen von MinMax Baumen
function build_subtree(i::Int, j::Int, is_max::Bool, branching_factor::Int, outcomes::Vector{Int}) :: MinMaxNode
	if j-i > branching_factor
		ranges = Int((j-i) / branching_factor)
		sub_nodes = [
			build_subtree(i+(b*ranges), i+((b+1)*ranges), !is_max, branching_factor, outcomes)
				for b in 0:branching_factor-1
		]
		return MinMaxNode(is_max, sub_nodes)
	else # Blaetter des Baumes erstellen
		return MinMaxNode(is_max, [
			MinMaxLeaf(o) for o in outcomes[i:j-1]
		])
	end
end
function build_tree(depth::Int, branching_factor::Int, outcomes::Vector{Int}) :: MinMaxTree
	needed_leafs = branching_factor ^ depth
	if length(outcomes) != needed_leafs
		throw(ArgumentError("Needs $needed_leafs outcomes for depth=$depth and branching_factor=$branching_factor !"))
	end

	root = build_subtree(1, needed_leafs+1, true, branching_factor, outcomes)	
	return MinMaxTree(root)
end

# Ausgabe von MinMax Baeumen
function Base.show(io::IO, t::MinMaxTree)
	println(io, "Min Max Tree:")
	if isnothing(t.root)
		println("\tEmpty Tree")
	else
		print(io, t.root)
	end
end
function Base.show(io::IO, s::MinMaxNode, indent::Int=0)
	print(io, repeat("   ", indent))
	printstyled(io, if s.is_max "<" else ">" end, bold=true)
	alpha = if typemin(Int) == s.alpha "-Inf" else s.alpha end
	beta = if typemax(Int) == s.beta "Inf" else s.beta end
	println( " -|", if isnothing(s.v) " (abgeschnitten)" else string(" ($(s.v), a=$(alpha), b=$(beta))") end)
	for su in s.subsequent_states
		show(io, su, indent+1)
	end
end
function Base.show(io::IO, s::MinMaxLeaf, indent::Int=0)
	print(io, repeat("   ", indent), "- ")
	printstyled(io, s.value, bold=true)
	println()
end

"""
	Alpha-Beta-Pruning mit Cutoff

	Foliensatz: 11-Suchraumbeschneidung
	Ca. ab Folie(n): 42
"""

is_terminal(state::MinMaxLeaf) = true
is_terminal(state::MinMaxNode) = false

function eval(state::MinMaxLeaf)
	return state.value
end
function eval(state::MinMaxNode)
	#  approximiere den Wert eines Zusatndes (fuer Spiel etc. **anpassen!**)
	#		hier wissen wir, dass nach einem Schritt die Blaetter und nehmen 
	#		dort den ersten Wert
	return state.subsequent_states[1].value
end
function cutoff_test(state::Union{MinMaxNode, MinMaxLeaf}, depth::Int)
	# Limit fuer die Tiefe der Rechung (Schritte in die Zukunft); stoppe an Blaettern
	return is_terminal(state) || depth > 2
end

function alpha_beta_search(t::MinMaxTree)
	alpha = typemin(Int)
	beta = typemax(Int)
	
	best = max_value(t.root, alpha, beta, 1)
	#	bestimme Aktion/ Zustand zu bestem Wert
	for (i, sub_state) in enumerate(t.root.subsequent_states)
		if best == sub_state.v
			return (i, sub_state)
		end
	end
end

function max_value(state::Union{MinMaxNode, MinMaxLeaf}, alpha::Int, beta::Int, depth::Int)
	if cutoff_test(state, depth)
		return eval(state)
	end

	v = typemin(Int)
	for sub_state in state.subsequent_states
		v = maximum((v, min_value(sub_state, alpha, beta, depth+1)))
		if v >= beta 
			break
		end 
		alpha = maximum((alpha, v))
	end
	state.alpha = alpha
	state.beta = beta
	state.v = v
	return v
end

function min_value(state::Union{MinMaxNode, MinMaxLeaf}, alpha::Int, beta::Int, depth::Int)
	if is_terminal(state)
		return eval(state)
	end

	v = typemax(Int)
	for sub_state in state.subsequent_states
		v = minimum((v, max_value(sub_state, alpha, beta, depth+1)))
		if v <= alpha 
			break
		end 
		beta = minimum((beta, v))
	end
	state.alpha = alpha
	state.beta = beta
	state.v = v
	return v
end

# Beispielaufrufe
t = build_tree(3, 3, [6, 5, 18, 6, -19, -18, -16, -13, -9, -17, 13, -20, -15, 2, 3, -1, 0, 0, -7, -19, -5, -19, 14, -8, -19, -8, -11])
(a_index, a_node) = alpha_beta_search(t)
println("Beste Aktion:\n\tAktion Nr. $a_index mit dem Wert $(a_node.v):")
println(t)