#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	The main graph implementation used in chapter 10.

		The implementation has to goal to provide all necessary features
		we need for the algorithms in chapter 10.
		*Thus, tis implementation is quite complex, is is not necessary to
		understand it. Examples about its usage can be found in
		`10-Graphen/a-graph-example.jl`.*

	> Please do not change the code here, otherwise some examples using this
	> code may fail!

"""

##
#	Global settings, ...
##
if !@isdefined(USE_GRAPH_PLOTS)
	USE_GRAPH_PLOTS = true
end
if USE_GRAPH_PLOTS && !haskey(ENV, "JULIA_AUD_PLOT_GRAPHS_DISABLE")
	try 
		using Plots, GraphRecipes
	catch 
		println("Install Packages GraphRecipes and Plots for plotted Graphs.")
		println("import Pkg; Pkg.add(\"Plots\"); Pkg.add(\"GraphRecipes\"); ")
	end
end

# Used for iterating on edges, nodes
include("./generator.jl")

##
#	Type
##
abstract type Graph end
mutable struct GraphStruct <: Graph
	name_node :: Dict{String, Any} # nodes by their names 
	node_name :: Dict{Any, String} # get name for node

	adjacency_in :: Dict{String, Dict{String, Number}} # edges pointing to `node` from `node_`: `[node][node_][weight]`
	adjacency_out :: Dict{String, Dict{String, Number}} # edges pointing from `node` to `node_`: `[node][node_][weight]`

	weighted :: Bool # is graph weighted
	directed :: Bool # is graph directed

	num_nodes :: Int # number of nodes 
	num_edges :: Int # number of edges

	name :: String
	plot_count :: Int 

	GraphStruct(w::Bool, d::Bool, name::String) = new(
		Dict{Any, String}(),
		Dict{String, Any}(),
		Dict{String, Dict{String, Number}}(),
		Dict{String, Dict{String, Number}}(),
		w, d,
		0, 0,
		name, 0
	)
end

##
#	Constructors
##
"""
	Create a Graph (unweighted and undirected).
	Each graph has a name, e.g., used for printing and plotting.

	Optionally give pairs of nodes by specifying edges `es` as 
	vector of pairs `u => v` (names will be inferred by `string(u)`). 
"""
Graph(name::String) :: Graph = Graph(Pair[], name) 
Graph(es::Vector{<:Pair}, name::String) :: Graph = Graph(es, Number[], name) :: Graph
Graph(es::Vector{<:Pair}, ws::Vector{<:Number}, name::String) :: Graph = Graph(es, ws, GraphStruct(false, false, name)) 
function Graph(edges::Vector{<:Pair}, weights::Vector{<:Number}, gs::GraphStruct) :: Graph
	if gs.weighted && (length(edges) != length(weights))
		throw(ArgumentError("Number of edges and weights need to be equal!"))
	end

	for (i, (u,v)) in enumerate(edges) 
		insert(u, gs) # same node with same name will replace node, if already added
		insert(v, gs)

		if gs.weighted
			insert(u, v, gs, weights[i])
		else 
			insert(u, v, gs)
		end
	end
	
	return gs
end

# More Constructors (weighted and directed graphs)
"""
	Create a weighted Graph (undirected).
	Each graph has a name, e.g., used for printing and plotting.

	Optionally give pairs of nodes by specifying edges `es` as 
	vector of pairs `u => v` (names will be inferred by `string(u)`). 
	Aditionally give vector of weights per edge/ pair.
"""
WeightedGraph(name::String) :: Graph = WeightedGraph(Pair[], Number[], name) 
WeightedGraph(es::Vector{<:Pair}, ws::Vector{<:Number}, name::String) :: Graph = Graph(es, ws, GraphStruct(true, false, name))

"""
	Create a directed Graph (unweighted).

	Optionally give pairs of nodes by specifying edges `es` as 
	vector of pairs `u => v` (names will be inferred by `string(u)`). 
"""
DirectedGraph(name::String) :: Graph = DirectedGraph(Pair[], name)
DirectedGraph(es::Vector{<:Pair}, name::String) :: Graph = Graph(es, Number[], GraphStruct(false, true, name))

"""
	Create a directed and weighted Graph.

	Optionally give pairs of nodes by specifying edges `es` as 
	vector of pairs `u => v` (names will be inferred by `string(u)`). 
	Additionally give vector of weights per edge/ pair.
"""
DirectedWeightedGraph(name::String) :: Graph = DirectedWeightedGraph(Pair[], Number[], name)
DirectedWeightedGraph(es::Vector{<:Pair}, ws::Vector{<:Number}, name::String) :: Graph = Graph(es, ws, GraphStruct(true, true, name))

##
#	Functions
##
"""
	Insert node `v` into the Graph `gs`.
	The nodes have names, which have to be unqiue, give name via `name`
	or `string(v)` is used.
"""
insert(v::Any, gs::Graph) = insert(v, gs, nothing)
function insert(v::Any, gs::Graph, name::Union{Nothing, String})
	# get name
	if isnothing(name)
		name = string(v)
	end

	# if name already taken (replace and delete old mapping `node => name`)
	if haskey(gs.name_node, name)
		delete!(gs.node_name, gs.name_node[name])
	else 
		# new node!
		gs.num_nodes += 1
	end

	# add node (lookup via name and node possible)
	gs.node_name[v] = name
	gs.name_node[name] = v

	# assure dicts in nodes adjacency dicts
	if !haskey(gs.adjacency_out, name)
		gs.adjacency_out[name] = Dict{String, Number}()
		gs.adjacency_in[name] = Dict{String, Number}()
	end
end

"""
	Insert an edge `uv` in graph `gs` specified as pair `u => v` or by two 
	parameters `u`, `v`.

	If graph is weighted, specify `weight`.
"""
insert(uv::Pair, gs::Graph) = insert(uv.first, uv.second, gs::Graph)
insert(uv::Pair, gs::Graph, weight::Number) = insert(uv.first, uv.second, gs::Graph, weight)
function insert(u::Any, v::Any, gs::Graph, weight::Union{Nothing, <:Number}=nothing)
	if gs.weighted && isnothing(weight)
		throw(ArgumentError("Give weight in weighted graphs!"))
	elseif !gs.weighted && !isnothing(weight)
		throw(ArgumentError("Do not give weight in unweighted graphs!"))
	end

	# retrieve names
	u = gs.node_name[u]
	v = gs.node_name[v]

	# default weight for unweighted graphs
	if !gs.weighted
		weight = 1
	end

	if (gs.directed && !haskey(gs.adjacency_out[u], v)) ||
		(!gs.directed && !haskey(gs.adjacency_out[v], u) && !haskey(gs.adjacency_out[u], v))
		# only new edge if not already created!
		gs.num_edges += 1
	end

	gs.adjacency_out[u][v] = weight
	gs.adjacency_in[v][u] = weight
	if !gs.directed
		gs.adjacency_out[v][u] = weight
		gs.adjacency_in[u][v] = weight
	end 
end

"""
	Check if a node `u` (by name or node) is in graph `gs`.
"""
function Base.in(u::Any, gs::Graph)
	return haskey(gs.node_name, u)
end
function Base.in(u::String, gs::Graph)
	return haskey(gs.name_node, u)
end

"""
	Find by node `name` in `gs`.

	Also possible to retrieve using `gs[name]`.
"""
function find(name::String, gs::Graph)
	return gs.name_node[name]
end
Base.getindex(gs::Graph, name::String) = find(name, gs)

"""
	Find edge by between `(u,v)` by nodes `u`, `v`,
	their names, or pair `u=>v` in `gs`.

	Will return `false` if no edge found and weight in weighted graphs,
	else `true`.

	Also possible to retrieve using `gs[u, v]`.
"""
find(u::Any, v::Any, gs::Graph) = find(gs.node_name[u], gs.node_name[v], gs)
find(uv::Pair, gs::Graph) = find(uv.first, uv.second, gs)
function find(u::String, v::String, gs::Graph)
	if haskey(gs.adjacency_out[u], v)
		if gs.weighted 
			return gs.adjacency_out[u][v]
		else 
			return true
		end
	else 
		return false 
	end
end
Base.getindex(gs::Graph, u::Any, v::Any) = find(u, v, gs)
Base.getindex(gs::Graph, uv::Pair) = find(uv, gs)

"""
	Remove node `u` by name or node in `gs`.

	Removes also all edges from and to node `u`.
"""
remove(u::Any, gs::Graph) = remove(gs.node_name[u], gs)
function remove(u::String, gs::Graph)
	# delete edges
	for (v, _) in edges_in(u, gs, true)
		remove(v, u, gs)
	end
	for (_, v) in edges_out(u, gs, true)
		remove(u, v, gs)
	end
	
	# delete adjacency dicts
	delete!(gs.adjacency_out, u)
	delete!(gs.adjacency_in, u)	

	# delete node
	delete!(gs.node_name, gs.name_node[u])
	delete!(gs.name_node, u)

	gs.num_nodes -= 1
end	

"""
	Remove edge by between `(u,v)` by nodes `u`, `v`,
	their names, or pair `u=>v` in `gs`.
"""
remove(u::Any, v::Any, gs::Graph) = remove(gs.node_name[u], gs.node_name[v], gs)
remove(uv::Pair, gs::Graph) = remove(uv.first, uv.second, gs)
function remove(u::String, v::String, gs::Graph)
	# delete edge
	delete!(gs.adjacency_out[u], v)
	delete!(gs.adjacency_in[v], u)

	if !gs.directed
		# other direction too, if undirected
		delete!(gs.adjacency_out[v], u)
		delete!(gs.adjacency_in[u], v)
	end

	gs.num_edges -= 1
end

"""
	Iterate all nodes in the graph `gs` (not ordered!)

	Gets node objects if `names==false`, else node names
"""
function nodes(gs::Graph, names::Bool=false)
	return if names keys(gs.name_node) else values(gs.name_node) end
end

"""
	Get the number of nodes in `g`.
"""
function num_nodes(g::GraphStruct)
	return g.num_nodes 
end

"""
	Iterate all edges of graph `gs`.

	Gets node objects if `names==false`, else node names.
	
	Each edge is a tuple of `(u,v)` or `(u,v,w)` in weighted 
	graphs.
"""
function edges(gs::Graph, names::Bool=false)
	return Generator(
		function (uvw)
			(u, v, w) = uvw
			if !names # get node objects for names
				u = gs.name_node[u]
				v = gs.name_node[v]
			end
			# only add weights in weighted graphs
			if gs.weighted
				return u, v, w
			else
				return u, v
			end
		end,
		(
			(u, v, w)
				for u in keys(gs.adjacency_out) 			# for each node's outgoing edges
					for (v,w) in pairs(gs.adjacency_out[u])	# and each weight there
		),
		(uvw) -> (gs.directed || uvw[1] < uvw[2]) # filter double edges (if graph undirected)
	)
end

"""
	Get the number of edges in `g`.
"""
function num_edges(g::GraphStruct)
	return g.num_edges
end

"""
	Getting all edges `(u,v)` going out of node `u` (either use name or node as `u`.)

	Gets node objects if `names==false`, else node names.
"""
edges_out(u::Any, gs::Graph, names::Bool=false) = edges_out(gs.node_name[u], gs, names)
function edges_out(u::String, gs::Graph, names::Bool=false)
	u_ret = if names u else gs.name_node[u] end
	return Generator(
		function (vw)
			(v, w) = vw
			if !names 
				v = gs.name_node[v]
			end
			if gs.weighted
				return u_ret, v, w
			else
				return u_ret, v
			end
		end,
		pairs(gs.adjacency_out[u])
	)
end

"""
	Getting all edges `(u,v)` going out in node `v` (either use name or node as `v`.)

	Gets node objects if `names==false`, else node names.
"""
edges_in(v::Any, gs::Graph, names::Bool=false) = edges_in(gs.node_name[v], gs, names)
function edges_in(v::String, gs::Graph, names::Bool=false)
	v_ret = if names v else gs.name_node[v] end
	return Generator(
		function (vw)
			(u, w) = vw
			if !names 
				u = gs.name_node[u]
			end
			if gs.weighted
				return u, v_ret, w
			else
				return u, v_ret
			end
		end,
		pairs(gs.adjacency_in[v])
	)
end

function Base.show(io::IO, gs::Graph)

	# get all node names
	names = Vector{String}(undef, gs.num_nodes)
	for (i,name) in enumerate(keys(gs.name_node))
		names[i] = name
	end
	sort!(names)

	# create adjacency matrix
	adjacency = zeros(Int8, (gs.num_nodes, gs.num_nodes))
	weights = zeros(Number, (gs.num_nodes, gs.num_nodes))
	for (i,u) in enumerate(names)
		for (j,v) in enumerate(names)
			if haskey(gs.adjacency_out[u], v)
				adjacency[i, j] = 1

				weight = gs.adjacency_out[u][v] 
				weights[i, j] = if isinteger(weight) convert(Int, weight) else weight end
			end
		end
	end

	if @isdefined(Plots) && @isdefined(GraphRecipes)
		# create plot, if libs loaded
		if gs.num_nodes > 0 && gs.num_edges > 0
			gs.plot_count += 1
			safe_name = replace(gs.name, r"[^A-Za-z0-9\_\-\.]" => s"")
			graph_hash = hash(weights) + hash(adjacency) + hash(names)

			plot_name = "./plot_$(safe_name)_$(gs.plot_count)_$(graph_hash).png"
			if !isfile(plot_name)
				fig = graphplot(
					adjacency, names=names,
					edgelabel=if gs.weighted weights else nothing end,
					node_shape=:circle, fontsize=12
				)
				savefig(fig, plot_name)
			end
		end
	end

	# name & type of graph
	if gs.directed && gs.weighted
		println(io, "Gerichteter und gewichteter Graph -- \"$(gs.name)\"")
	elseif gs.directed
		println(io, "Gerichteter Graph -- \"$(gs.name)\"")
	elseif gs.weighted 
		println(io, "Gewichteter Graph -- \"$(gs.name)\"")
	else 
		println(io, "\nGraph (ungerichtetet, ungewichtetet) -- \"$(gs.name)\"")
	end

	# list of nodes
	println(io, "\tKnoten: $(join(names, ", "))")

	# plot the adjacency matrix on terminal
	PAD_SIZE = 4
	print(io, "\tAdjazenzmatrix:\n\t", ' '^(PAD_SIZE+4))
	for (_,v) in enumerate(names)
		print(io, lpad(v, PAD_SIZE, ' '), ' ')
	end
	println(io, "\n\t", ' '^(PAD_SIZE+3), '-'^(gs.num_nodes*(PAD_SIZE+1)))
	for (i,u) in enumerate(names)
		print(io, '\t', lpad(u, PAD_SIZE+1, ' '), " : " )
		for (j,_) in enumerate(names)
			if adjacency[i, j] == 1
				print(io, lpad(weights[i, j], PAD_SIZE, ' '), ' ')
			else 
				print(io, ' '^(PAD_SIZE+1))
			end
		end 
		println(io)
	end

	# more stats
	println(io, "\tAnzahl Knoten: $(gs.num_nodes), Anzahl Kanten: $(gs.num_edges)")
end
